# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import warnings
from enum import Enum
from types import GenericAlias
from typing import Any, Literal, cast

import shortuuid
from lancedb.pydantic import LanceModel
from pydantic import create_model
from typing_extensions import TYPE_CHECKING

from pixano.datasets.queries import TableQueryBuilder
from pixano.datasets.utils.errors import DatasetIntegrityError
from pixano.features import (
    AnnotationRef,
    BaseSchema,
    EmbeddingRef,
    EntityRef,
    SchemaGroup,
    SchemaRef,
    Source,
    ViewRef,
    is_annotation_ref,
    is_embedding_ref,
    is_entity_ref,
    is_item_ref,
    is_schema_ref,
    is_source_ref,
    is_view_ref,
)


if TYPE_CHECKING:
    from pixano.datasets import Dataset


class IntegrityCheck(Enum):
    """Integrity check types.

    Attributes:
        DEFINED_ID: Check if the id field is defined.
        UNIQUE_ID: Check if the id field is unique.
        REF_NAME: Check if the ref name is defined in the schema.
        REF_TYPE: Check if the ref type is defined in the schema.
        REF_ID: Check if the ref id is stored in the referenced table.
    """

    DEFINED_ID = 0
    UNIQUE_ID = 1
    REF_NAME = 2
    REF_TYPE = 3
    REF_ID = 4


def get_integry_checks_from_schemas(
    schemas: list[BaseSchema], table_name: str
) -> list[list[tuple[str, str, str, str, Any]]]:
    """Get the integrity checks to perform on a table.

    Args:
        schemas: List of schemas to check.
        table_name: Table name.

    Returns:
        List of integrity checks to perform on the table. The checks are grouped by type.
        - check_id: Check id (unique identifier for the checks). It is used to avoid checking subsequent checks with
            the same id when an error is found.
        - table: Table name.
        - schema_id: Schema id which is the id field value from the schema.
        - field_name: Field name to check.
        - field: Field value to check.
    """
    checks: list[list[tuple[str, str, str, str, Any]]] = [[] for _ in IntegrityCheck]
    for schema in schemas:
        schema_id = schema.id
        check_id = shortuuid.uuid()
        checks[IntegrityCheck.DEFINED_ID.value].append((check_id, table_name, schema_id, "id", schema_id))
        checks[IntegrityCheck.UNIQUE_ID.value].append((check_id, table_name, schema_id, "id", schema_id))
        for field_name, field in schema.model_fields.items():
            if field_name == "id":
                continue
            if isinstance(field.annotation, GenericAlias):
                continue
            type_field = field.annotation
            if is_schema_ref(type_field):
                checks[IntegrityCheck.REF_NAME.value].append(
                    (check_id, table_name, schema_id, field_name, getattr(schema, field_name))
                )
                checks[IntegrityCheck.REF_TYPE.value].append(
                    (check_id, table_name, schema_id, field_name, getattr(schema, field_name))
                )
                checks[IntegrityCheck.REF_ID.value].append(
                    (check_id, table_name, schema_id, field_name, getattr(schema, field_name))
                )

    return checks


def check_table_integrity(
    table_name: str,
    dataset: "Dataset",
    schemas: list[BaseSchema] | None = None,
    updating: bool = False,
    ignore_checks: list[IntegrityCheck] | None = None,
) -> list[tuple[IntegrityCheck, str, str, str, Any]]:
    """Check the integrity of schemas against a table.

    Args:
        table_name: Table name.
        dataset: Dataset that contains the table.
        schemas: List of schemas to insert in table. If None, the table is checked, otherwise the schemas are checked
            against the table.
        updating: If True, the table is being updated. It is used to avoid checking the id uniqueness when updating
            schemas.
        ignore_checks: List of integrity checks to ignore.

    Returns:
        List of errors as tuples with the following values:
        - check_type: Check type.
        - table: Table name.
        - field_name: Field name that caused the error.
        - schema_id: Schema id that raised the error.
        - field: Field value that caused the error.
    """
    table = dataset.open_table(table_name)

    if ignore_checks is not None:
        ignore_checks_set: set[IntegrityCheck] = {IntegrityCheck(check) for check in ignore_checks}
    else:
        ignore_checks_set = set()
    table_schema = Source if table_name == "source" else dataset.schema.schemas[table_name]

    checking_table = schemas is None
    if schemas is None:
        if updating:
            raise ValueError("schemas must be provided when updating a table.")
        table_schema = cast(BaseSchema, table_schema)
        fields_to_check = ["id"] + [
            field_name
            for field_name, field in table_schema.model_fields.items()
            if field_name != "id"
            and not isinstance(field.annotation, GenericAlias)
            and is_schema_ref(field.annotation)
        ]
        model = create_model(
            "_Schema",
            __base__=LanceModel,
            **{field_name: (table_schema.model_fields[field_name].annotation, ...) for field_name in fields_to_check},
        )
        schemas = TableQueryBuilder(table).select(fields_to_check).to_pydantic(model)

    table_ids = [schema.id for schema in schemas]
    count_ids: dict[str, int] = {}
    for id in table_ids:
        count_ids[id] = count_ids.get(id, 0) + 1
    integrity_checks = get_integry_checks_from_schemas(schemas, table_name)
    check_errors: dict[str, tuple[IntegrityCheck, str, str, str, Any]] = {}
    ids_to_check: dict[str, str] = {}
    schemas_refs_to_check: dict[str, list[tuple[str, str, SchemaRef, str]]] = {}

    for check_type_id, checks in enumerate(integrity_checks):
        check_type = IntegrityCheck(check_type_id)
        if check_type in ignore_checks_set:
            continue
        for check_id, _, schema_id, field_name, field in checks:
            if check_id in check_errors:
                continue
            if check_type == IntegrityCheck.DEFINED_ID and field == "":  # id is not defined
                check_errors[check_id] = (check_type, table_name, field_name, schema_id, field)
            elif check_type == IntegrityCheck.UNIQUE_ID:
                if count_ids[schema_id] > 1:  # id is not unique
                    check_errors[check_id] = (check_type, table_name, field_name, schema_id, field)
                elif not checking_table:
                    ids_to_check[schema_id] = check_id
            elif check_type == IntegrityCheck.REF_NAME:
                field = cast(SchemaRef, field)
                if field.name != "" and field.name not in (
                    list(dataset.schema.schemas.keys()) + ["source"]
                ):  # ref name is not defined
                    check_errors[check_id] = (check_type, table_name, field_name, schema_id, field)
            elif check_type == IntegrityCheck.REF_TYPE:
                field = cast(SchemaRef, field)
                if field.name == "":
                    continue
                field_type = type(field)
                if is_view_ref(field_type):  # field is a view ref
                    field = cast(ViewRef, field)
                    if field.name not in dataset.schema.groups[SchemaGroup.VIEW]:  # field name is not a view
                        check_errors[check_id] = (check_type, table_name, field_name, schema_id, field)
                elif is_annotation_ref(field_type):  # field is an annotation ref
                    field = cast(AnnotationRef, field)
                    if (
                        field.name not in dataset.schema.groups[SchemaGroup.ANNOTATION]
                    ):  # field name is not an annotation
                        check_errors[check_id] = (check_type, table_name, field_name, schema_id, field)
                elif is_embedding_ref(field_type):  # field is an embedding ref
                    field = cast(EmbeddingRef, field)
                    if (
                        field.name not in dataset.schema.groups[SchemaGroup.EMBEDDING]
                    ):  # field name is not an embedding
                        check_errors[check_id] = (check_type, table_name, field_name, schema_id, field)
                elif is_entity_ref(field_type):  # field is an entity ref
                    field = cast(EntityRef, field)
                    if field.name not in dataset.schema.groups[SchemaGroup.ENTITY]:  # field name is not an entity
                        check_errors[check_id] = (check_type, table_name, field_name, schema_id, field)
                elif is_item_ref(field_type) or is_source_ref(field_type):
                    pass  # item_ref and source_ref are validated before.
            elif check_type == IntegrityCheck.REF_ID:  # ref id and ref item relation checked below
                field = cast(SchemaRef, field)
                if field_name == "":
                    continue
                # If the field is empty, the reference is to the table itself so no need to check
                if field.id == "":
                    continue
                if field.name not in schemas_refs_to_check:
                    schemas_refs_to_check[field.name] = []
                schemas_refs_to_check[field.name].append((check_id, schema_id, field, field_name))

    if not checking_table and not updating and len(ids_to_check) > 0:
        for id, found in dataset.find_ids_in_table(table_name, set(ids_to_check.keys())).items():
            if found:
                check_errors[ids_to_check[id]] = (IntegrityCheck.UNIQUE_ID, table_name, "id", id, id)

    if len(check_errors) == len(
        {check_id for check_id, *_ in integrity_checks[IntegrityCheck.REF_ID.value]}
    ):  # all checks failed, no need to check later checks that are costly
        return list(check_errors.values())

    for ref_schema_name, refs in schemas_refs_to_check.items():
        if ref_schema_name == "":
            continue
        ref_ids_to_check = {field_ref.id for check_id, _, field_ref, _ in refs if check_id not in check_errors}
        found_ref_ids = dataset.find_ids_in_table(ref_schema_name, ref_ids_to_check)
        for check_id, schema_id, field_ref, field_name in refs:
            if check_id in check_errors:
                continue
            if not found_ref_ids[field_ref.id]:
                check_errors[check_id] = (
                    IntegrityCheck.REF_ID,
                    table_name,
                    field_name,
                    schema_id,
                    field_ref,
                )

    return list(check_errors.values())


def check_dataset_integrity(dataset: "Dataset") -> list[tuple[IntegrityCheck, str, str, str, Any]]:
    """Check the integrity of a dataset.

    Args:
        dataset: Dataset to check.

    Returns:
        List of errors as tuples with the following values:
        - check_type: Check type.
        - table: Table name.
        - field_name: Field name that caused the error.
        - schema_id: Schema id that raised the error.
        - field: Field value that caused the error.
    """
    check_errors: list[tuple[IntegrityCheck, str, str, str, Any]] = []
    for table_name in dataset.schema.schemas.keys():
        check_errors.extend(check_table_integrity(table_name, dataset))
    return check_errors


def handle_integrity_errors(
    check_errors: list[tuple[IntegrityCheck, str, str, str, Any]],
    raise_or_warn: Literal["raise", "warn"] = "raise",
) -> None:
    """Handle integrity check errors.

    Args:
        check_errors: List of errors.
        raise_or_warn: If "raise", raise a ValueError with the errors. If "warn", warns a UserWarning with the errors.
    """
    if len(check_errors) == 0:
        return
    message = "Integrity check errors:\n"
    for check_type, table_name, field_name, schema_id, field in check_errors:
        message += "- "
        if check_type == IntegrityCheck.DEFINED_ID:
            message += f"An id is not defined in table {table_name}.\n"
        elif check_type == IntegrityCheck.UNIQUE_ID:
            message += f"The id {schema_id} is not unique in table {table_name}.\n"
        elif check_type == IntegrityCheck.REF_NAME:
            message += f"The reference {field_name} from {schema_id} to the table {field.name} does not exist.\n"
        elif check_type == IntegrityCheck.REF_TYPE:
            message += (
                f"The reference {field_name} from {schema_id} to the table {field.name} is to an invalid type. "
                f"Got {type(field)}.\n"
            )
        elif check_type == IntegrityCheck.REF_ID:
            message += (
                f"The reference {field_name} from {schema_id} to the table {field.name} has an invalid id. Got "
                f"{field.id}.\n"
            )
    if raise_or_warn == "raise":
        raise DatasetIntegrityError(message)
    else:
        warnings.warn(message, category=UserWarning)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pydantic import BaseModel, ConfigDict, field_validator

from pixano.features.schemas.registry import _PIXANO_SCHEMA_REGISTRY
from pixano.features.schemas.schema_group import SchemaGroup


class TableInfo(BaseModel):
    """The information of a table.

    Attributes:
        name: Table name.
        group: Group of the schema associated with the table.
        base_schema: Base Pixano schema stored in the registry.
    """

    model_config = ConfigDict(validate_assignment=True)
    name: str
    group: str
    base_schema: str

    @field_validator("base_schema")
    @classmethod
    def _must_be_registered(cls, value: str) -> str:
        """Check that the base schema is registered."""
        if value not in _PIXANO_SCHEMA_REGISTRY:
            raise ValueError(f"Schema {value} is not registered.")
        return value

    @field_validator("group")
    @classmethod
    def _must_be_group(cls, value: str) -> str:
        """Check that the group is valid."""
        if not any(value == group.value for group in SchemaGroup):
            raise ValueError(f"Group {value} is not valid.")
        return value



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from .conversation import Conversation, create_conversation, is_conversation
from .entity import Entity, is_entity
from .track import Track, create_track, is_track


__all__ = [
    "Conversation",
    "Entity",
    "Track",
    "create_conversation",
    "create_track",
    "is_conversation",
    "is_entity",
    "is_track",
]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import json
from pathlib import Path

from pydantic import BaseModel


class DatasetFeaturesValues(BaseModel):
    """Constraints for the dataset features values.

    Attributes:
        items: Constraints for the dataset item table.
        views: Constraints for the dataset view tables.
        entities: Constraints for the dataset entity tables.
        annotations: Constraints for the dataset annotation tables.
    """

    items: dict[str, list] = {}
    views: dict[str, list] = {}
    entities: dict[str, list] = {}
    annotations: dict[str, list] = {}

    def to_json(self, json_fp: Path) -> None:
        """Save DatasetFeaturesValues to json file."""
        json_fp.write_text(json.dumps(self.model_dump(), indent=4), encoding="utf-8")

    @staticmethod
    def from_json(json_fp: Path) -> "DatasetFeaturesValues":
        """Load DatasetFeaturesValues from json file."""
        fv_json = json.loads(json_fp.read_text(encoding="utf-8"))
        fv = DatasetFeaturesValues.model_validate(fv_json)

        return fv



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import pyarrow as pa


SERIALIZE_PYARROW_DATATYPE = {
    pa.float16(): "halffloat",
    pa.float32(): "float",
    pa.float64(): "double",
    pa.int8(): "int8",
    pa.int16(): "int16",
    pa.int32(): "int32",
    pa.int64(): "int64",
    pa.uint8(): "uint8",
    pa.uint16(): "uint16",
    pa.uint32(): "uint32",
    pa.uint64(): "uint64",
    pa.binary(): "binary",
    pa.bool_(): "bool",
}

DESERIALIZE_PYARROW_DATATYPE = {v: k for k, v in SERIALIZE_PYARROW_DATATYPE.items()}



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from typing import Annotated

from fastapi import APIRouter, Depends, Query

from pixano.app.models.annotations import AnnotationModel
from pixano.app.settings import Settings, get_settings
from pixano.features.schemas.schema_group import SchemaGroup

from .utils import (
    create_row_handler,
    create_rows_handler,
    delete_row_handler,
    delete_rows_handler,
    get_row_handler,
    get_rows_handler,
    update_row_handler,
    update_rows_handler,
)


router = APIRouter(prefix="/annotations", tags=["Annotations"])


@router.get("/{dataset_id}/{table}/", response_model=list[AnnotationModel])
async def get_annotations(
    dataset_id: str,
    table: str,
    settings: Annotated[Settings, Depends(get_settings)],
    ids: list[str] | None = Query(None),
    limit: int | None = None,
    skip: int = 0,
    where: str | None = None,
    item_ids: list[str] | None = Query(None),
) -> list[AnnotationModel]:
    """Get annotations from a table of a dataset.

    They can be filtered by IDs, item IDs, a where clause or paginated.

    Args:
        dataset_id: Dataset ID containing the table.
        table: Table name.
        settings: App settings.
        ids: IDs of the annotations.
        limit: Limit number of annotations.
        skip: Skip number of annotations.
        where: Where clause.
        item_ids: Item IDs.

    Returns:
        List of annotations.
    """
    return await get_rows_handler(
        dataset_id=dataset_id,
        group=SchemaGroup.ANNOTATION,
        table=table,
        settings=settings,
        where=where,
        ids=ids,
        item_ids=item_ids,
        limit=limit,
        skip=skip,
    )


@router.get("/{dataset_id}/{table}/{id}", response_model=AnnotationModel)
async def get_annotation(
    dataset_id: str, table: str, id: str, settings: Annotated[Settings, Depends(get_settings)]
) -> AnnotationModel:
    """Get an annotation from a table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        table: Table name.
        id: ID of the annotation.
        settings: App settings.

    Returns:
        The annotation.
    """
    return await get_row_handler(dataset_id, SchemaGroup.ANNOTATION, table, id, settings)


@router.post("/{dataset_id}/{table}/", response_model=list[AnnotationModel])
async def create_annotations(
    dataset_id: str,
    table: str,
    annotations: list[AnnotationModel],
    settings: Annotated[Settings, Depends(get_settings)],
) -> list[AnnotationModel]:
    """Add annotations in a table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        table: Table name.
        annotations: Annotations to add.
        settings: App settings.

    Returns:
        List of annotations added.
    """
    return await create_rows_handler(dataset_id, SchemaGroup.ANNOTATION, table, annotations, settings)


@router.post("/{dataset_id}/{table}/{id}", response_model=AnnotationModel)
async def create_annotation(
    dataset_id: str,
    table: str,
    id: str,
    annotation: AnnotationModel,
    settings: Annotated[Settings, Depends(get_settings)],
) -> AnnotationModel:
    """Add an annotation in a table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        table: Table name.
        id: ID of the annotation.
        annotation: Annotation to add.
        settings: App settings.

    Returns:
        The annotation added.
    """
    return await create_row_handler(dataset_id, SchemaGroup.ANNOTATION, table, id, annotation, settings)


@router.put("/{dataset_id}/{table}/{id}", response_model=AnnotationModel)
async def update_annotation(
    dataset_id: str,
    table: str,
    id: str,
    annotation: AnnotationModel,
    settings: Annotated[Settings, Depends(get_settings)],
) -> AnnotationModel:
    """Update an annotation in a table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        table: Table name.
        id: ID of the annotation.
        annotation: Annotation to update.
        settings: App settings.

    Returns:
        The annotation updated.
    """
    return await update_row_handler(dataset_id, SchemaGroup.ANNOTATION, table, id, annotation, settings)


@router.put("/{dataset_id}/{table}/", response_model=list[AnnotationModel])
async def update_annotations(
    dataset_id: str,
    table: str,
    annotations: list[AnnotationModel],
    settings: Annotated[Settings, Depends(get_settings)],
) -> list[AnnotationModel]:
    """Update annotations in a table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        table: Table name.
        annotations: Annotations to update.
        settings: App settings.

    Returns:
        List of annotations updated.
    """
    return await update_rows_handler(dataset_id, SchemaGroup.ANNOTATION, table, annotations, settings)


@router.delete("/{dataset_id}/{table}/{id}")
async def delete_annotation(
    dataset_id: str, table: str, id: str, settings: Annotated[Settings, Depends(get_settings)]
) -> None:
    """Delete an annotation from a table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        table: Table name.
        id: ID of the annotation to delete.
        settings: App settings.
    """
    return await delete_row_handler(dataset_id, SchemaGroup.ANNOTATION, table, id, settings)


@router.delete("/{dataset_id}/{table}/")
async def delete_annotations(
    dataset_id: str,
    table: str,
    ids: Annotated[list[str], Query()],
    settings: Annotated[Settings, Depends(get_settings)],
) -> None:
    """Delete annotations from a table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        table: Table name.
        ids: IDs of the annotations to delete.
        settings: App settings.
    """
    return await delete_rows_handler(dataset_id, SchemaGroup.ANNOTATION, table, ids, settings)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import json

from pydantic import ConfigDict, field_validator
from typing_extensions import Self

from pixano.datasets import Dataset
from pixano.features import SchemaGroup, Source

from .base_schema import BaseSchemaModel
from .table_info import TableInfo


class SourceModel(BaseSchemaModel[Source]):
    """Model for the [Source][pixano.features.Source] schema."""

    model_config = ConfigDict(
        validate_assignment=True,
        json_schema_extra={
            "examples": [
                {
                    "id": "source_0",
                    "table_info": {"group": "source", "name": "source", "base_schema": "Source"},
                    "data": {
                        "name": "source_0",
                        "kind": "model",
                        "metadata": {"model_id": "model_0"},
                    },
                }
            ]
        },
    )

    @field_validator("table_info")
    @classmethod
    def _validate_table_info(cls, value: TableInfo) -> TableInfo:
        """Validate table info."""
        if value.group != SchemaGroup.SOURCE.value:
            raise ValueError(f"Table info group must be {SchemaGroup.SOURCE.value}.")
        return value

    def to_row(self, dataset: Dataset) -> Source:
        """Create a [Source][pixano.features.Source] from the model."""
        schema_dict = self.model_dump()
        row = Source.model_validate(
            {
                "id": schema_dict["id"],
                "created_at": schema_dict["created_at"],
                "updated_at": schema_dict["updated_at"],
                **schema_dict["data"],
            }
        )
        row.dataset = dataset
        row.table_name = self.table_info.name
        row.metadata = json.dumps(self.data["metadata"])
        return row

    @classmethod
    def from_row(cls, row: Source, table_info: TableInfo) -> Self:
        """Create a SourceModel from a Source.

        Args:
            row: The row to create the model from.
            table_info: The table info of the row.

        Returns:
            The created model.
        """
        source_model = BaseSchemaModel.from_row(row, table_info)
        source_model.data["metadata"] = json.loads(source_model.data["metadata"])
        return cls.model_construct(**source_model.__dict__)  # Avoid validation and casting



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from .table import TableQueryBuilder


__all__ = ["TableQueryBuilder"]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pixano.utils import issubclass_strict

from ..base_schema import BaseSchema
from ..registry import _register_schema_internal


@_register_schema_internal
class Item(BaseSchema):
    """Item base class.

    Items are used to store information about an item in a dataset. It contains at least a split attribute.
    It also federates the information about the item's views, entities, annotations, embeddings, etc via its id.

    Attributes:
        split: Split of the item.
    """

    split: str = "default"


def is_item(cls: type, strict: bool = False) -> bool:
    """Check if a class is an `Item` or subclass of `Item`."""
    return issubclass_strict(cls, Item, strict)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from typing import Annotated

from fastapi import APIRouter, Depends, HTTPException, Query

from pixano.app.models.dataset_items import DatasetItemModel
from pixano.app.settings import Settings, get_settings
from pixano.datasets.dataset import DatasetAccessError, DatasetPaginationError

from .utils import get_dataset


router = APIRouter(prefix="/dataset_items", tags=["Dataset Items"])


@router.get("/{dataset_id}/", response_model=list[DatasetItemModel])
async def get_dataset_items(
    dataset_id: str,
    settings: Annotated[Settings, Depends(get_settings)],
    ids: list[str] | None = Query(None),
    limit: int | None = None,
    skip: int = 0,
) -> list[DatasetItemModel]:
    """Get dataset items from a dataset.

    Args:
        dataset_id: Dataset ID containing the items.
        settings: App settings.
        ids: IDs of the dataset items.
        limit: Limit number of dataset items.
        skip: Skip number of dataset items.

    Returns:
        List of dataset items.
    """
    dataset = get_dataset(dataset_id, settings.library_dir, None)

    try:
        rows = dataset.get_dataset_items(ids, limit, skip)
    except DatasetPaginationError as err:
        raise HTTPException(status_code=400, detail="Invalid query parameters. " + str(err))
    except DatasetAccessError as err:
        raise HTTPException(status_code=500, detail="Insternal server error. " + str(err))
    if rows == []:
        raise HTTPException(status_code=404, detail="Dataset items not found.")

    return DatasetItemModel.from_dataset_items(rows, dataset.schema)


@router.get("/{dataset_id}/{id}", response_model=DatasetItemModel)
async def get_dataset_item(
    dataset_id: str, id: str, settings: Annotated[Settings, Depends(get_settings)]
) -> DatasetItemModel:
    """Get dataset item from a dataset.

    Args:
        dataset_id: Dataset ID.
        id: Dataset item ID.
        settings: App settings.

    Returns:
        The dataset item.
    """
    return (await get_dataset_items(dataset_id, settings, ids=[id]))[0]


@router.post("/{dataset_id}/", response_model=list[DatasetItemModel])
async def create_dataset_items(
    dataset_id: str,
    dataset_items: list[DatasetItemModel],
    settings: Annotated[Settings, Depends(get_settings)],
) -> list[DatasetItemModel]:
    """Add dataset items in a dataset.

    Args:
        dataset_id: Dataset ID to add the items.
        dataset_items: Dataset items to add.
        settings: App settings.

    Returns:
        List of dataset items added.
    """
    dataset = get_dataset(dataset_id, settings.library_dir, None)

    try:
        rows = dataset.add_dataset_items(DatasetItemModel.to_dataset_items(dataset_items, dataset))
    except ValueError:
        raise HTTPException(status_code=400, detail="Invalid dataset items")
    return DatasetItemModel.from_dataset_items(rows, dataset.schema)


@router.post("/{dataset_id}/{id}", response_model=DatasetItemModel)
async def create_dataset_item(
    dataset_id: str,
    id: str,
    dataset_item: DatasetItemModel,
    settings: Annotated[Settings, Depends(get_settings)],
) -> DatasetItemModel:
    """Add a dataset item.

    Args:
        dataset_id: Dataset ID to add the item.
        id: Dataset item ID to add.
        dataset_item: Dataset item.
        settings: App settings.

    Returns:
        The dataset item added.
    """
    if id != dataset_item.id:
        raise HTTPException(status_code=400, detail="ID in path and body do not match.")

    return (await create_dataset_items(dataset_id, [dataset_item], settings))[0]


@router.put("/{dataset_id}/", response_model=list[DatasetItemModel])
async def update_dataset_items(
    dataset_id: str,
    dataset_items: list[DatasetItemModel],
    settings: Annotated[Settings, Depends(get_settings)],
) -> list[DatasetItemModel]:
    """Update dataset items in a dataset.

    Args:
        dataset_id: Dataset ID containing the items.
        dataset_items: Dataset items to update.
        settings: App settings.

    Returns:
        List of dataset items updated.
    """
    dataset = get_dataset(dataset_id, settings.library_dir, None)

    try:
        rows = dataset.update_dataset_items(DatasetItemModel.to_dataset_items(dataset_items, dataset))
    except ValueError:
        raise HTTPException(status_code=400, detail="Invalid dataset items")
    return DatasetItemModel.from_dataset_items(rows, dataset.schema)


@router.put("/{dataset_id}/{id}", response_model=DatasetItemModel)
async def update_dataset_item(
    dataset_id: str,
    id: str,
    dataset_item: DatasetItemModel,
    settings: Annotated[Settings, Depends(get_settings)],
) -> DatasetItemModel:
    """Update dataset item in a dataset.

    Args:
        dataset_id: Dataset ID containing the item.
        id: Dataset item ID to update.
        dataset_item: Dataset item.
        settings: App settings.

    Returns:
        The dataset item updated.
    """
    if id != dataset_item.id:
        raise HTTPException(status_code=400, detail="ID in path and body do not match.")

    return (await update_dataset_items(dataset_id, [dataset_item], settings))[0]


@router.delete("/{dataset_id}/")
async def delete_dataset_items(
    dataset_id: str,
    ids: Annotated[list[str], Query()],
    settings: Annotated[Settings, Depends(get_settings)],
) -> None:
    """Delete dataset items from a dataset.

    Args:
        dataset_id: Dataset ID containing the items.
        ids: IDs of the dataset items to delete.
        settings: App settings.
    """
    dataset = get_dataset(dataset_id, settings.library_dir, None)

    dataset.delete_dataset_items(ids)
    return


@router.delete("/{dataset_id}/{id}")
async def delete_dataset_item(dataset_id: str, id: str, settings: Annotated[Settings, Depends(get_settings)]) -> None:
    """Delete a dataset item from a dataset.

    Args:
        dataset_id: Dataset ID containing the item.
        id: ID of the dataset item to delete.
        settings: App settings.
    """
    return await delete_dataset_items(dataset_id, ids=[id], settings=settings)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pathlib import Path
from typing import Annotated

from fastapi import APIRouter, Depends, HTTPException

from pixano.app.models import DatasetModel
from pixano.app.models.dataset_info import DatasetInfoModel
from pixano.app.settings import Settings, get_settings
from pixano.datasets import DatasetInfo
from pixano.datasets.utils.errors import DatasetAccessError

from .utils import get_dataset as get_dataset_utils


router = APIRouter(prefix="/datasets", tags=["Datasets"])


@router.get("/info", response_model=list[DatasetInfoModel])
async def get_datasets_info(
    settings: Annotated[Settings, Depends(get_settings)],
) -> list[DatasetInfoModel]:
    """Load a list of dataset information.

    Args:
        settings: App settings.

    Returns:
        List of dataset info.
    """
    try:
        infos_and_paths: list[tuple[DatasetInfo, Path]] = DatasetInfo.load_directory(
            directory=settings.library_dir, return_path=True
        )
    except FileNotFoundError:
        raise HTTPException(
            status_code=404,
            detail=f"No datasets found in {settings.library_dir.absolute()}.",
        )

    if len(infos_and_paths) > 0:
        return [DatasetInfoModel.from_dataset_info(info, path) for info, path in infos_and_paths]
    raise HTTPException(
        status_code=404,
        detail=f"No datasets found in {settings.library_dir.absolute()}.",
    )


@router.get("/info/{id}", response_model=DatasetInfoModel)
async def get_dataset_info(
    id: str,
    settings: Annotated[Settings, Depends(get_settings)],
) -> DatasetInfoModel:
    """Load a single dataset information.

    Args:
        id: Dataset ID to load info from.
        settings: App settings.

    Returns:
        The dataset info.
    """
    try:
        info, path = DatasetInfo.load_id(id, settings.library_dir, return_path=True)
    except FileNotFoundError:
        raise HTTPException(
            status_code=404,
            detail=f"Dataset {id} not found in {settings.library_dir.absolute()}.",
        )

    return DatasetInfoModel.from_dataset_info(info, path)


@router.get("/{id}", response_model=DatasetModel)
async def get_dataset(
    id: str,
    settings: Annotated[Settings, Depends(get_settings)],
) -> DatasetModel:
    """Load dataset from ID.

    Args:
        id: Dataset ID to load.
        settings: App settings.

    Returns:
        Dataset model.
    """
    return DatasetModel.from_dataset(get_dataset_utils(id, settings.library_dir, settings.media_dir))


@router.get("/{id}/{table}/count", response_model=int)
async def get_table_count(
    id: str,
    table: str,
    settings: Annotated[Settings, Depends(get_settings)],
) -> int:
    """Get the number of rows in a table.

    Args:
        id: Dataset ID containing the table.
        table: Table name.
        settings: App settings.

    Returns:
        The number of rows in the table.
    """
    dataset = get_dataset_utils(id, settings.library_dir, settings.media_dir)
    try:
        db_table = dataset.open_table(table)
    except DatasetAccessError as e:
        raise HTTPException(
            status_code=404,
            detail=str(e),
        )
    return db_table.count_rows()



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import re
from dataclasses import dataclass
from datetime import datetime
from typing import Literal

from pydantic import model_validator

from pixano.utils import issubclass_strict

from ...types.schema_reference import EntityRef, ItemRef, SourceRef, ViewRef
from ..registry import _register_schema_internal
from .annotation import Annotation


@dataclass
class Answer:
    """Holds the parts of an answer for multi-choices questions.

    Raises:
        ValueError: in case of parsing error at initialisation
    """

    content: str
    _choices: list[str]
    _explanation: str

    @classmethod
    def parse(cls, content: str):
        """Utility function that parses a str content into the 3 fields of the Answer object.

        The input should be formated as follow "[[A;B;C]] this is a justification."

        Returns:
            content, the reformated content
            choices, the list of parsed answer elements
            explanation, the explanation sentence
        """
        choices = []
        explanation = ""
        try:
            re_choices_explanation = r"\[\[([a-zA-Z0-9]+(\;[a-zA-Z0-9]+)*)\]\]\s*(.*)"
            match_1 = re.match(re_choices_explanation, content)
            if match_1:
                explanation = str(match_1.groups()[-1])
                match_2 = re.findall(r"[a-zA-Z0-9]+", match_1.groups()[0])
                for m in match_2:
                    choices.append(m)
            content = Answer.format(choices, explanation)
        except Exception:
            raise ValueError(f"could not parse answer message : {content} => {choices}, {explanation}")
        return content, choices, explanation

    @classmethod
    def format(cls, choices, explanation):
        """Formats the message content for frontend.

        Returns:
            str: formated message content
        """
        return "[[{0}]] {1}".format(";".join(choices), explanation)

    def __init__(self, content: str):
        """Constructor of the Answer from a lazy formated message.

        Exemple of content : "[[A;B;C]] this is a justification."

        Args:
            content (str): lazy formated content of the message.
        """
        self.content, self._choices, self._explanation = Answer.parse(content)

    def __str__(self):
        """Formats the message content for frontend.

        Returns:
            str: formated message content
        """
        return Answer.format(self._choices, self._explanation)


@_register_schema_internal
class Message(Annotation):
    """Textual exchange in a question/answer conversation
    for image or text description and information extraction.

    Attributes:
        number: message number to associate different ANSWER messages to a QUESTION.
        content: actual text of the message.
        user: identify who is the author of the message (eg a human, a model, the ground truth, etc).
        choices: list of allowed answers, for 'Multiple Choice Question' when type=QUESTION and question_type!=OPEN.
        timestamp: creation date of the message.
        type: type of the message within "SYSTEM", "QUESTION" or "ANSWER".
            - SYSTEM: used for prefix messages stating the context. No associated answer expected
            - QUESTION: used to ask a question about a View. Expecting at least one answer (same message number)
            - ANSWER: used to reply to a question message by refering its message number
        question_type: type of question, specifying how to read and parse the content field.
            Authorized calued within "OPEN", "SINGLE_CHOICE", "SINGLE_CHOICE_EXPLANATION",
                                     "MULTI_CHOICE", "MULTI_CHOICE_EXPLANATION".
            - OPEN: used for any open question, where no specific format of answer is expected
            - SINGLE_CHOICE: used for a multi-choice-question where only one answer is authorized.
            - SINGLE_CHOICE_EXPLANATION: similar to SINGLE_CHOICE, but an explanation is expected
                    after the choosen answer. Cf Answer object
            - MULTI_CHOICE: used for a multi-choice-question where multi answers are authorized.
            - MULTI_CHOICE_EXPLANATION: similar to MULTI_CHOICE, but an explanation is expected
                    after the choosen answers. Cf Answer object
    """

    content: str
    number: int
    user: str
    type: str
    question_type: str = "OPEN"
    choices: list[str] = []
    timestamp: datetime = datetime(1, 1, 1, 0, 0, 0, 0)

    @model_validator(mode="before")
    def _check_data(data):
        if "type" not in data:
            raise ValueError("the message expect a type, either QUESTION or ANSWER.")
        data["type"] = data.get("type").upper()
        data["question_type"] = data.get("question_type", "OPEN").upper()
        data["choices"] = data.get("choices", [])
        if data["type"] == "QUESTION":
            if data["question_type"] == "OPEN" and len(data["choices"]) > 0:
                data["question_type"] = "MULTI_CHOICE"
            elif data["question_type"] != "OPEN" and len(data["choices"]) == 0:
                data["question_type"] = "OPEN"

        if data["type"] == "ANSWER":
            if data["question_type"] != "OPEN":
                answer = Answer(data["content"])
                if "_EXPLANATION" in data["question_type"] and answer._explanation == "":
                    raise ValueError("the question type expect an explanation in the answer.")
                if "SINGLE_" in data["question_type"] and len(answer._choices) > 1:
                    raise ValueError("the question type expect only one selected choice in the answer.")
                data["content"] = str(answer)
        return data

    @model_validator(mode="after")
    def _validate_fields(self):
        if self.number < 0:
            raise ValueError("number should be a positive or null integer")
        elif self.type not in ["SYSTEM", "QUESTION", "ANSWER"]:
            raise ValueError("Message type should be 'SYSTEM', 'QUESTION', or 'ANSWER'.")
        if self.type != "QUESTION" and len(self.choices) > 0:
            raise ValueError(
                f"specifying MCQ anwsers is only allowed for Message of type 'QUESTION'. type={self.type}."
            )
        if self.type == "QUESTION" and self.question_type not in [
            "OPEN",
            "SINGLE_CHOICE",
            "SINGLE_CHOICE_EXPLANATION",
            "MULTI_CHOICE",
            "MULTI_CHOICE_EXPLANATION",
        ]:
            raise ValueError(
                "Question type should be 'OPEN',"
                " 'SINGLE_CHOICE', 'SINGLE_CHOICE_EXPLANATION',"
                " 'MULTI_CHOICE', or 'MULTI_CHOICE_EXPLANATION'."
            )
        if self.type == "QUESTION" and self.question_type != "OPEN" and not self.choices:
            raise ValueError(f"MCQ should have a list of available choices: {self}")
        return self

    @classmethod
    def none(cls) -> "Message":
        """Utility function to get a None equivalent.
        Should be removed when Lance could manage None value.

        Returns:
            "None" Message.
        """
        return cls(
            id="",
            item=ItemRef.none(),
            view=ViewRef.none(),
            entity=EntityRef.none(),
            source_ref=SourceRef.none(),
            number=0,
            user="",
            type="QUESTION",
            content="",
            choices=[],
            timestamp=datetime(1, 1, 1, 0, 0, 0, 0),
        )


def is_message(cls: type, strict: bool = False) -> bool:
    """Check if a class is a Message or subclass of Message."""
    return issubclass_strict(cls, Message, strict)


def create_message(
    number: int,
    user: str,
    type: Literal["SYSTEM", "QUESTION", "ANSWER"],
    content: str,
    choices: list[str] = [],
    timestamp: datetime = datetime(1, 1, 1, 0, 0, 0, 0),
    id: str = "",
    item_ref: ItemRef = ItemRef.none(),
    view_ref: ViewRef = ViewRef.none(),
    entity_ref: EntityRef = EntityRef.none(),
    source_ref: SourceRef = SourceRef.none(),
) -> Message:
    """Create a Message instance.

    Args:
        number: message number to associate diffrent ANSWER messages to a QUESTION
        user: identify who is the author of the message (eg a human, a model, the ground truth, etc)
        type: type of the message within "SYSTEM", "QUESTION" or"ANSWER"
        content: actual text of the message
        choices: list of allowed answers
        timestamp: creation date of the message
        id: object id
        item_ref: Item reference.
        view_ref: View reference.
        entity_ref: Entity reference.
        source_ref: Source reference.

    Returns:
        The created `Message` instance.
    """
    return Message(
        number=number,
        user=user,
        type=type,
        content=content,
        choices=choices,
        timestamp=timestamp,
        id=id,
        item_ref=item_ref,
        view_ref=view_ref,
        entity_ref=entity_ref,
        source_ref=source_ref,
    )



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from .dataset import Dataset
from .dataset_features_values import DatasetFeaturesValues
from .dataset_info import DatasetInfo
from .dataset_schema import DatasetItem, DatasetSchema
from .dataset_stat import DatasetStatistic
from .queries import TableQueryBuilder
from .workspaces import DefaultImageDatasetItem, DefaultVideoDatasetItem, DefaultVQADatasetItem, WorkspaceType


__all__ = [
    "Dataset",
    "DatasetFeaturesValues",
    "DatasetInfo",
    "DatasetItem",
    "DatasetSchema",
    "DatasetStatistic",
    "DefaultImageDatasetItem",
    "DefaultVideoDatasetItem",
    "DefaultVQADatasetItem",
    "TableQueryBuilder",
    "WorkspaceType",
]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================


import json
from typing import Any

from fastapi.encoders import jsonable_encoder

from pixano.features import (
    BaseSchema,
    SchemaGroup,
    Source,
    group_to_str,
    schema_to_group,
)

from ..dataset_info import DatasetInfo
from ..dataset_schema import DatasetItem
from .dataset_exporter import DatasetExporter


class DefaultJSONDatasetExporter(DatasetExporter):
    """Default JSON dataset exporter."""

    def initialize_export_data(self, info: DatasetInfo, sources: list[Source]) -> dict[str, Any]:
        """Initialize the dictionary or list of dictionaries to be exported.

        Args:
            info: The dataset information.
            sources: The list of sources.

        Returns:
            A dictionary containing the data to be exported.
        """
        export_data = {"info": info.model_dump()}

        for group, schemas in self.dataset.schema.groups.items():
            if group == SchemaGroup.EMBEDDING:
                continue
            elif group == SchemaGroup.ITEM:
                export_data[group_to_str(group, plural=True)] = []
            else:
                export_data[group_to_str(group, plural=True)] = {schema: [] for schema in schemas}
        export_data[group_to_str(SchemaGroup.SOURCE, plural=True)] = [
            s.model_dump(exclude_timestamps=True) for s in sources
        ]
        return export_data

    def export_dataset_item(self, export_data: dict[str, Any], dataset_item: DatasetItem) -> dict[str, Any]:
        """Store the dataset item in the `export_data` dictionary.

        Args:
            export_data: A dictionary containing the data to be exported.
            dataset_item: The dataset item to be exported.

        Returns:
            A dictionary containing the data to be exported.
        """
        data: dict[str, BaseSchema | list[BaseSchema] | None] = dataset_item.to_schemas_data(self.dataset.schema)
        for schema_name, schema_data in data.items():
            if schema_data:
                schema_data = schema_data if isinstance(schema_data, list) else [schema_data]
                group = schema_to_group(schema_data[0])
                if group == SchemaGroup.ITEM:
                    export_data[group_to_str(group, plural=True)].extend(
                        [s.model_dump(exclude_timestamps=True) for s in schema_data]
                    )
                else:
                    export_data[group_to_str(group, plural=True)][schema_name].extend(
                        [s.model_dump(exclude_timestamps=True) for s in schema_data]
                    )
        return export_data

    def save_data(self, export_data: dict[str, Any], split: str, file_name: str, file_num: int) -> None:
        """Save data to the specified directory.

        The saved directory has the following structure:
            export_dir/{split}_{file_name}_0.json
                      /...
                      /{split}_{file_name}_{file_num}.json
                      /...
                      /{split}_{file_name}_n.json


        Args:
            export_data: The dictionary containing the data to be saved.
            split: The split of the dataset item being saved.
            file_name: The name of the file to save the data in.
            file_num: The number of the file to save the data in.
        """
        json_path = self.export_dir / f"{split}_{file_name}_{file_num}.json"
        json_path.write_text(json.dumps(jsonable_encoder(export_data), indent=4), encoding="utf-8")



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import numpy as np

from .image import rle_to_mask, urle_to_rle


def denormalize_coords(coord: list[float], height: int, width: int, rounded_int: bool = True) -> list[float]:
    """Denormalize coordinates.

    Args:
        coord: Normalized coordinates.
        height: Height.
        width: Width.
        rounded_int: True to round denormalized float to nearest integer.

    Returns:
        Unnormalized coordinates.
    """
    denorm = []

    for i, c in enumerate(coord):
        if i % 2 == 0:
            denorm.append(round(c * width) if rounded_int else c * width)
        else:
            denorm.append(round(c * height) if rounded_int else c * height)

    return denorm


def normalize_coords(coord: list[float], height: int, width: int) -> list[float]:
    """Normalize coordinates.

    Args:
        coord: Unnormalized coordinates.
        height: Height.
        width: Width.

    Returns:
        Normalized coordinates.
    """
    norm = []

    for i, c in enumerate(coord):
        if i % 2 == 0:
            norm.append(c / width)
        else:
            norm.append(c / height)

    return norm


def mask_to_bbox(mask: np.ndarray) -> list[float]:
    """Return the smallest bounding box containing all the mask pixels.

    Args:
        mask: Mask as NumPy Array.

    Returns:
        Normalized xywh bounding box.
    """
    height, width = mask.shape
    bool_mask = np.array(mask).astype(bool)

    # Find all columns and rows that contain ones
    rows = np.any(bool_mask, axis=1)
    cols = np.any(bool_mask, axis=0)

    # Find the min and max col/row index that contain ones
    rmin, rmax = np.where(rows)[0][[0, -1]]
    cmin, cmax = np.where(cols)[0][[0, -1]]

    # Calculate bbox height and width
    w = (cmax - cmin + 1) / width
    h = (rmax - rmin + 1) / height

    return [cmin / width, rmin / height, w, h]


def urle_to_bbox(urle: dict) -> list[float]:
    """Return the smallest bounding box containing all the mask pixels.

    Args:
        urle: Mask as uncompressed RLE.

    Returns:
        Normalized xywh bounding box.
    """
    return mask_to_bbox(rle_to_mask(urle_to_rle(urle)))


def xywh_to_xyxy(xywh: list[float]) -> list[float]:
    """Convert bounding box coordinates from xywh
    (using top left point as reference) to xyxy.

    Args:
        xywh: xywh coordinates.

    Returns:
        xyxy coordinates.
    """
    return [
        float(xywh[0]),
        float(xywh[1]),
        float(xywh[0] + xywh[2]),
        float(xywh[1] + xywh[3]),
    ]


def xyxy_to_xywh(xyxy: list[float]) -> list[float]:
    """Convert bounding box coordinates from xyxy to xywh
    (using top left point as reference).

    Args:
        xyxy: xyxy coordinates.

    Returns:
        xywh coordinates.
    """
    return [
        float(xyxy[0]),
        float(xyxy[1]),
        float(xyxy[2] - xyxy[0]),
        float(xyxy[3] - xyxy[1]),
    ]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from .annotation import Annotation, is_annotation
from .bbox import BBox, BBox3D, create_bbox, create_bbox3d, is_bbox, is_bbox3d
from .camcalibration import (
    BaseIntrinsics,
    CamCalibration,
    Extrinsics,
    Intrinsics,
    create_cam_calibration,
    is_cam_calibration,
)
from .classification import Classification, create_classification, is_classification
from .compressed_rle import CompressedRLE, create_compressed_rle, is_compressed_rle
from .info_extraction import Relation, TextSpan, create_relation, create_text_span, is_relation, is_text_span
from .keypoints import KeyPoints, KeyPoints3D, create_keypoints, create_keypoints3d, is_keypoints, is_keypoints3d
from .text_generation import Message, create_message, is_message
from .tracklet import Tracklet, create_tracklet, is_tracklet


__all__ = [
    "Annotation",
    "BaseIntrinsics",
    "BBox",
    "BBox3D",
    "CamCalibration",
    "Classification",
    "CompressedRLE",
    "Extrinsics",
    "Intrinsics",
    "KeyPoints",
    "KeyPoints3D",
    "Message",
    "TextSpan",
    "Tracklet",
    "Relation",
    "is_annotation",
    "is_bbox",
    "is_bbox3d",
    "is_cam_calibration",
    "is_classification",
    "is_compressed_rle",
    "is_keypoints",
    "is_keypoints3d",
    "is_message",
    "is_text_span",
    "is_tracklet",
    "is_relation",
    "create_bbox",
    "create_bbox3d",
    "create_cam_calibration",
    "create_classification",
    "create_compressed_rle",
    "create_keypoints",
    "create_keypoints3d",
    "create_message",
    "create_text_span",
    "create_tracklet",
    "create_relation",
]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from typing import Annotated

from fastapi import APIRouter, Depends, Query

from pixano.app.models.sources import SourceModel
from pixano.app.settings import Settings, get_settings
from pixano.features.schemas.schema_group import SchemaGroup

from .utils import (
    create_row_handler,
    create_rows_handler,
    delete_row_handler,
    delete_rows_handler,
    get_row_handler,
    get_rows_handler,
    update_row_handler,
    update_rows_handler,
)


router = APIRouter(prefix="/sources", tags=["Sources"])


@router.get("/{dataset_id}/", response_model=list[SourceModel])
async def get_sources(
    dataset_id: str,
    settings: Annotated[Settings, Depends(get_settings)],
    ids: list[str] | None = Query(None),
    limit: int | None = None,
    skip: int = 0,
    where: str | None = None,
) -> list[SourceModel]:
    """Get sources from the `'source'` table of a dataset.

    They can be filtered by IDs, a where clause or paginated.

    Args:
        dataset_id: Dataset ID containing the table.
        settings: App settings.
        ids: IDs of the sources.
        limit: Limit number of sources.
        skip: Skip number of sources.
        where: Where clause.

    Returns:
        List of sources.
    """
    return await get_rows_handler(
        dataset_id=dataset_id,
        group=SchemaGroup.SOURCE,
        table=SchemaGroup.SOURCE.value,
        settings=settings,
        where=where,
        ids=ids,
        item_ids=None,
        limit=limit,
        skip=skip,
    )


@router.get("/{dataset_id}/{id}", response_model=SourceModel)
async def get_source(dataset_id: str, id: str, settings: Annotated[Settings, Depends(get_settings)]) -> SourceModel:
    """Get a source from the `'source'` table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        id: ID of the source.
        settings: App settings.

    Returns:
        The source.
    """
    return await get_row_handler(dataset_id, SchemaGroup.SOURCE, SchemaGroup.SOURCE.value, id, settings)


@router.post("/{dataset_id}/", response_model=list[SourceModel])
async def create_sources(
    dataset_id: str,
    sources: list[SourceModel],
    settings: Annotated[Settings, Depends(get_settings)],
) -> list[SourceModel]:
    """Add sources in the `'source'` table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        sources: Sources to add.
        settings: App settings.

    Returns:
        List of sources added.
    """
    return await create_rows_handler(dataset_id, SchemaGroup.SOURCE, SchemaGroup.SOURCE.value, sources, settings)


@router.post("/{dataset_id}/{id}", response_model=SourceModel)
async def create_source(
    dataset_id: str,
    id: str,
    source: SourceModel,
    settings: Annotated[Settings, Depends(get_settings)],
) -> SourceModel:
    """Add a source in the `'source'` table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        id: ID of the source.
        source: Source to add.
        settings: App settings.

    Returns:
        The source added.
    """
    return await create_row_handler(dataset_id, SchemaGroup.SOURCE, SchemaGroup.SOURCE.value, id, source, settings)


@router.put("/{dataset_id}/{id}", response_model=SourceModel)
async def update_source(
    dataset_id: str,
    id: str,
    source: SourceModel,
    settings: Annotated[Settings, Depends(get_settings)],
) -> SourceModel:
    """Update a source in the `'source'` table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        id: ID of the source.
        source: Source to update.
        settings: App settings.

    Returns:
        The source updated.
    """
    return await update_row_handler(dataset_id, SchemaGroup.SOURCE, SchemaGroup.SOURCE.value, id, source, settings)


@router.put("/{dataset_id}/", response_model=list[SourceModel])
async def update_sources(
    dataset_id: str,
    sources: list[SourceModel],
    settings: Annotated[Settings, Depends(get_settings)],
) -> list[SourceModel]:
    """Update sources in the `'source'` table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        sources: Sources to update.
        settings: App settings.

    Returns:
        List of sources updated.
    """
    return await update_rows_handler(dataset_id, SchemaGroup.SOURCE, SchemaGroup.SOURCE.value, sources, settings)


@router.delete("/{dataset_id}/{id}")
async def delete_source(dataset_id: str, id: str, settings: Annotated[Settings, Depends(get_settings)]) -> None:
    """Delete a source from the `'source'` table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        id: ID of the source to delete.
        settings: App settings.
    """
    return await delete_row_handler(dataset_id, SchemaGroup.SOURCE, SchemaGroup.SOURCE.value, id, settings)


@router.delete("/{dataset_id}/")
async def delete_sources(
    dataset_id: str,
    ids: Annotated[list[str], Query()],
    settings: Annotated[Settings, Depends(get_settings)],
) -> None:
    """Delete sources from the `'source'` table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        ids: IDs of the sources to delete.
        settings: App settings.
    """
    return await delete_rows_handler(dataset_id, SchemaGroup.SOURCE, SchemaGroup.SOURCE.value, ids, settings)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import math
import os
import os.path
from pathlib import Path

from PIL import Image, ImageDraw, ImageFont


def generate_mosaic_name(image_files, extension=".jpg"):
    """Generate a name for mosaic file whime keeping most common path / filename as possible."""
    if not image_files:
        return "mosaic" + extension
    common_path = os.path.commonpath(image_files)
    filenames = [os.path.basename(f) for f in image_files]
    common_prefix = os.path.commonprefix(filenames).rstrip("_- ")
    mosaic_name = f"{common_prefix}_mosaic{extension}" if common_prefix else f"mosaic{extension}"
    return os.path.join(common_path, mosaic_name)


def arrange_grid(num_images):
    """Get optimal grid distribution (rows, cols) for num_images."""
    cols = math.ceil(math.sqrt(num_images))
    rows = math.ceil(num_images / cols)
    return rows, cols


def compute_average_size(images):
    """Compute images average size."""
    widths, heights = zip(*(img.size for img in images))
    avg_width = sum(widths) // len(images)
    avg_height = sum(heights) // len(images)
    return avg_width, avg_height


def add_label_above(image, text, label_height=30):
    """Add a label at top left of image."""
    width, height = image.size

    label_img = Image.new("RGB", (width, label_height), (0, 0, 0))  # Fond noir
    draw = ImageDraw.Draw(label_img)

    # get Arial font, else system font
    try:
        font = ImageFont.truetype("arial.ttf", 20)
    except IOError:
        font = ImageFont.load_default()

    # Center text
    text_size = draw.textbbox((0, 0), text, font=font)  # Mesurer le texte
    text_width = text_size[2] - text_size[0]
    text_height = text_size[3] - text_size[1]

    text_x = (width - text_width) // 2
    text_y = (label_height - text_height) // 2

    # Draw text (white)
    draw.text((text_x, text_y), text, fill=(255, 255, 255), font=font)

    # Concat label and original image
    new_img = Image.new("RGB", (width, height + label_height))
    new_img.paste(label_img, (0, 0))  # Add label over
    new_img.paste(image, (0, label_height))  # put image under

    return new_img


def create_mosaic(
    source_dir: Path, split, image_files, output_path, label_prefix="image", padding=10, label_height=30
):
    """Create a mosaic from image_files with black background."""
    if not image_files:
        return

    need_split = False
    images = []
    for f in image_files:
        try:
            image = Image.open(source_dir / f)
        except FileNotFoundError:
            image = Image.open(source_dir / split / f)
            need_split = True
        image = image.convert("RGBA")
        images.append(image)

    rows, cols = arrange_grid(len(images))
    avg_width, avg_height = compute_average_size(images)

    images_resized = []
    for idx, img in enumerate(images):
        img_resized = img.resize((avg_width, avg_height), Image.LANCZOS)
        img_rgb = Image.new("RGB", img_resized.size, (0, 0, 0))
        img_rgb.paste(img_resized, mask=img_resized.split()[3])  # Apply alpha channel
        # Add label
        label_text = f"{label_prefix} {idx + 1}"
        img_labeled = add_label_above(img_rgb, label_text, label_height)
        images_resized.append(img_labeled)

    mosaic_width = cols * (avg_width + padding) - padding
    mosaic_height = rows * (avg_height + label_height + padding) - padding
    mosaic = Image.new("RGB", (mosaic_width, mosaic_height), (0, 0, 0))

    for idx, img in enumerate(images_resized):
        x_offset = (idx % cols) * (avg_width + padding)
        y_offset = (idx // cols) * (avg_height + label_height + padding)
        mosaic.paste(img, (x_offset, y_offset))

    if need_split:
        mosaic.save(source_dir / split / output_path, format="JPEG", quality=85, optimize=True)
    else:
        mosaic.save(source_dir / output_path, format="JPEG", quality=85, optimize=True)


def mosaic(source_dir: Path, split: str, image_files: list[str], view_name: str) -> str:
    """Create a mosaic from input images."""
    mosaic_filename = generate_mosaic_name(image_files)
    create_mosaic(source_dir, split, image_files, mosaic_filename, label_prefix=view_name, label_height=30, padding=5)
    return mosaic_filename



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pydantic import ConfigDict

from .items import ItemModel


class ItemInfoModel(ItemModel):
    """Item information.

    It contains all the information contained in an [ItemModel][pixano.app.models.ItemModel] and additional
    information about the dataset item such as the number of annotations, embeddings, entities and views.

    Attributes:
        infos: Information about the dataset item. Structure:
            {info_name: {sub_info_name: {"count": int, ...}, ...}, ...}
    """

    model_config = ConfigDict(
        validate_assignment=True,
        json_schema_extra={
            "examples": [
                {
                    "id": "1",
                    "table_info": {"group": "item", "name": "item", "base_schema": "Item"},
                    "data": {"split": "train", "source": "source1"},
                    "infos": {
                        "annotations": {"bbox": {"count": 0}, "keypoints": {"count": 0}},
                        "embeddings": {"image_embedding": {"count": 1}, "video_embedding": {"count": 2}},
                        "entities": {"persons": 0, "tracks": {"count": 10}},
                        "views": {"image": {"count": 1}, "video": {"count": 2}},
                    },
                }
            ]
        },
    )

    info: dict[str, dict[str, dict[str, int | float]]]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from typing import TYPE_CHECKING, Generic, TypeVar, Union

from pydantic import ConfigDict, model_validator

from pixano.utils import issubclass_strict

from . import BaseType
from .registry import _register_type_internal


if TYPE_CHECKING:
    from pixano.datasets import Dataset
    from pixano.features import Annotation, BaseSchema, Embedding, Entity, Item, Source, View


T = TypeVar("T", bound=Union["BaseSchema", "Item", "View", "Entity", "Annotation", "Embedding", "Source"])


class SchemaRef(BaseType, Generic[T]):
    """A schema reference.

    A schema reference is used to reference a schema in a dataset. If an id is provided, the reference points to a
    specific element stored in the table associated to the schema.

    Attributes:
        name: The name of the schema.
        id: The id of the schema.
    """

    model_config = ConfigDict(validate_assignment=True)
    name: str
    id: str = ""

    @classmethod
    def none(cls) -> "SchemaRef":
        """Return a reference to no schema."""
        return cls(name="", id="")

    def resolve(self, dataset: "Dataset") -> T:
        """Resolve the reference to the schema."""
        return dataset.resolve_ref(self)


@_register_type_internal
class ItemRef(SchemaRef["Item"]):
    """Item reference."""

    name: str = "item"  # hard coded to avoid circular import

    @model_validator(mode="after")
    def _validate_fields(self):
        if self.name != "item" and (  # hard coded to avoid circular import
            self.name != "" or self.id != ""
        ):
            raise ValueError("name must be 'item' when not empty.")
        return self


@_register_type_internal
class ViewRef(SchemaRef["View"]):
    """View reference."""

    pass


@_register_type_internal
class EntityRef(SchemaRef["Entity"]):
    """Entity reference."""

    pass


@_register_type_internal
class AnnotationRef(SchemaRef["Annotation"]):
    """Annotation reference."""

    pass


@_register_type_internal
class EmbeddingRef(SchemaRef["Embedding"]):
    """Embedding reference."""

    pass


@_register_type_internal
class SourceRef(SchemaRef["Source"]):
    """Source reference."""

    name: str = "source"  # hard coded to avoid circular import

    @model_validator(mode="after")
    def _validate_fields(self):
        if self.name != "source" and (  # hard coded to avoid circular import
            self.name != "" or self.id != ""
        ):
            raise ValueError("name must be 'source' when not empty.")
        return self


def is_schema_ref(cls: type, strict: bool = False) -> bool:
    """Check if a class is a `SchemaRef` or subclass of `SchemaRef`."""
    return issubclass_strict(cls, SchemaRef, strict)


def is_item_ref(cls: type, strict: bool = False) -> bool:
    """Check if a class is an `ItemRef` or subclass of `ItemRef`."""
    return issubclass_strict(cls, ItemRef, strict)


def is_view_ref(cls: type, strict: bool = False) -> bool:
    """Check if a class is a `ViewRef` or subclass of `ViewRef`."""
    return issubclass_strict(cls, ViewRef, strict)


def is_entity_ref(cls: type, strict: bool = False) -> bool:
    """Check if a class is an `EntityRef` or subclass of `EntityRef`."""
    return issubclass_strict(cls, EntityRef, strict)


def is_annotation_ref(cls: type, strict: bool = False) -> bool:
    """Check if a class is an `AnnotationRef` or subclass of `AnnotationRef`."""
    return issubclass_strict(cls, AnnotationRef, strict)


def is_embedding_ref(cls: type, strict: bool = False) -> bool:
    """Check if a class is an `EmbeddingRef` or subclass of `EmbeddingRef`."""
    return issubclass_strict(cls, EmbeddingRef, strict)


def is_source_ref(cls: type, strict: bool = False) -> bool:
    """Check if a class is a `SourceRef` or subclass of `SourceRef`."""
    return issubclass_strict(cls, SourceRef, strict)


def create_schema_ref(id: str, name: str) -> SchemaRef:
    """Create a schema reference."""
    return SchemaRef(name=name, id=id)


def create_item_ref(id: str, name: str = "item") -> ItemRef:
    """Create an item reference."""
    return ItemRef(name=name, id=id)


def create_view_ref(id: str, name: str) -> ViewRef:
    """Create a view reference."""
    return ViewRef(name=name, id=id)


def create_entity_ref(id: str, name: str) -> EntityRef:
    """Create an entity reference."""
    return EntityRef(name=name, id=id)


def create_annotation_ref(id: str, name: str) -> AnnotationRef:
    """Create an annotation reference."""
    return AnnotationRef(name=name, id=id)


def create_embedding_ref(id: str, name: str) -> EmbeddingRef:
    """Create an embedding reference."""
    return EmbeddingRef(name=name, id=id)


def create_source_ref(id: str) -> SourceRef:
    """Create a source reference."""
    return SourceRef(id=id)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================


import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any

import duckdb
import pyarrow as pa
import tqdm

from pixano.features import (
    SchemaGroup,
    Source,
)

from ..dataset import Dataset
from ..dataset_info import DatasetInfo
from ..dataset_schema import DatasetItem
from ..queries import TableQueryBuilder


logger = logging.getLogger(__name__)


class DatasetExporter(ABC):
    """Abstract base class for dataset exporters.

    To export a dataset, you need to implement this class and provide
    an implementation for the abstract methods.

    Attributes:
        dataset: The dataset to be exported.
        export_dir: The directory where the exported files will be saved.
    """

    def __init__(self, dataset: Dataset, export_dir: str | Path, overwrite: bool = False):
        """Initialize a new instance of the DatasetExporter class.

        Args:
            dataset: The dataset to be exported.
            export_dir: The directory where the exported files will be saved.
            overwrite: Whether to overwrite existing directory.
        """
        self.dataset = dataset
        self.export_dir = Path(export_dir)
        self._overwrite = overwrite

    @abstractmethod
    def initialize_export_data(self, info: DatasetInfo, sources: list[Source]) -> Any:
        """Initialize the data structure to be exported.

        Args:
            info: The dataset information.
            sources: The list of sources.

        Returns:
            The data structure to be exported.
        """
        ...

    @abstractmethod
    def export_dataset_item(self, export_data: Any, dataset_item: DatasetItem) -> Any:
        """Store the dataset item in the data structure to be exported.

        Args:
            export_data: The data structure to be exported.
            dataset_item: The dataset item to be exported.

        Returns: The data structure to be exported.
        """
        ...

    @abstractmethod
    def save_data(self, export_data: Any, split: str, file_name: str, file_num: int) -> None:
        """Save data to the specified directory.

        Args:
            export_data: The data structure to be exported.
            split: The split of the dataset item being saved.
            file_name: The name of the file to save the data in.
            file_num: The number of the file to save the data in.
        """
        ...

    def export(
        self, file_name: str = "pixano_export", items_per_file: int | None = None, batch_size: int | None = None
    ) -> None:
        """Export the dataset to the specified directory.

        Args:
            file_name: The name of the exported dataset.
            items_per_file: The number of items to export per file. If not specified, all items will be exported in a
                single file.
            batch_size: The size of each batch when exporting data. If not specified, all data will be exported at
                once.
        """
        self.export_dir.mkdir(exist_ok=self._overwrite)

        if items_per_file is not None and (not isinstance(items_per_file, int) or items_per_file <= 0):
            raise ValueError("Items per file must be a positive integer")
        if batch_size is not None and (not isinstance(batch_size, int) or batch_size <= 0):
            raise ValueError("Batch size must be a positive integer")

        info = self.dataset.info
        sources = self.dataset.get_data(SchemaGroup.SOURCE.value, limit=int(1e9))

        item_table = self.dataset.open_table(SchemaGroup.ITEM.value)
        item_table_lance = item_table.to_lance()  # noqa: F841
        SQL_QUERY = "SELECT split, COUNT(*) FROM item_table_lance GROUP BY split"
        arrow_results: pa.Table = duckdb.query(SQL_QUERY).to_arrow_table()
        splits: dict[str, list[Any]] = arrow_results.to_pydict()

        logger.info(f"Exporting dataset {self.dataset.info.name} to {self.export_dir}.")
        for num_split_items, split in zip(splits["count_star()"], splits["split"]):
            if items_per_file is None:
                split_items_per_file = num_split_items
            else:
                split_items_per_file = items_per_file
            if batch_size is None:
                batch_size = num_split_items

            export_data = self.initialize_export_data(info, sources)

            file_num = 0  # Number of files exported so far
            cur_items_exported = 0  # Number of items exported so far
            logger.info(
                f"Exporting split {split} of dataset {self.dataset.info.name}: number of items {num_split_items}"
            )

            for _ in tqdm.tqdm(
                range(0, num_split_items, batch_size),
                desc=f"Exporting data split {split} of dataset {self.dataset.info.name}",
            ):
                item_ids = list(
                    TableQueryBuilder(item_table)
                    .select("id")
                    .where(f"split='{split}'")
                    .limit(batch_size)
                    .offset(cur_items_exported)
                    .to_polars()["id"]
                )
                dataset_items: list[DatasetItem] = self.dataset.get_dataset_items(ids=item_ids)
                for dataset_item in dataset_items:
                    cur_items_exported += 1

                    export_data = self.export_dataset_item(export_data, dataset_item)

                    if (
                        cur_items_exported == num_split_items or cur_items_exported % split_items_per_file == 0
                    ):  # Export every n items
                        self.save_data(export_data, split, file_name, file_num)

                        file_num += 1
                        if cur_items_exported != num_split_items:
                            export_data = self.initialize_export_data(info, sources)

            logger.info(
                f"Completed export split {split} of dataset {self.dataset.info.name} in {file_num} file"
                f"{'s' if file_num > 1 else ''}."
            )



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import json
from abc import ABC
from math import prod
from typing import TYPE_CHECKING, Any, cast

import pyarrow as pa
from lancedb.embeddings import EmbeddingFunction
from lancedb.embeddings.registry import get_registry, register
from lancedb.pydantic import Vector
from pydantic import create_model, model_validator
from typing_extensions import Self

from pixano.features.schemas.views.image import Image, is_image
from pixano.features.schemas.views.sequence_frame import is_sequence_frame
from pixano.utils import issubclass_strict

from ...types.schema_reference import ItemRef, ViewRef
from ..base_schema import BaseSchema
from ..registry import _register_schema_internal


if TYPE_CHECKING:
    from pixano.datasets.dataset import Dataset
    from pixano.features.schemas import Item, View


def _to_pixano_name(dataset: "Dataset", table_name: str, name: str) -> str:
    return f"pixano_{dataset.info.id}_{table_name}_{name}"


def _from_pixano_name(dataset: "Dataset", table_name: str, pixano_name: str) -> str:
    return pixano_name[len(f"pixano_{dataset.info.id}_{table_name}_") :]


@_register_schema_internal
class Embedding(BaseSchema, ABC):
    """Embeddings are used to define an embedding vector for an item in a dataset.

    Attributes:
        item_ref: Reference to the embedding's item.
        vector: The embedding vector that should be defined by subclasses.
    """

    item_ref: ItemRef = ItemRef.none()
    vector: Any  # TODO: change to Vector exposed parametrized type when LanceDB is updated
    shape: list[int] = []

    @model_validator(mode="after")
    def _validate_vector(self) -> Self:
        if self.shape == []:
            self.shape = [len(self.vector)]
        if prod(self.shape) != len(self.vector):
            raise ValueError("The vector shape does not match the specified shape.")
        return self

    @property
    def item(self) -> "Item":
        """Get the embedding's item."""
        return self.resolve_ref(self.item_ref)

    @classmethod
    def to_arrow_schema(
        cls,
        remove_vector: bool = False,
        remove_metadata: bool = False,
    ) -> pa.Schema:
        """Get the pyarrow schema of an `Embedding`.

        This function allows to remove the vector field and the metadata from the schema which can be useful for adding
        data with auto-vectorization.

        Args:
            remove_vector: Remove the vector field.
            remove_metadata: Remove the metadata.

        Returns:
            The pyarrow schema.
        """
        pa_schema = super().to_arrow_schema()
        if remove_vector:
            pa_schema = pa_schema.remove(pa_schema.get_field_index("vector"))
        if remove_metadata:
            pa_schema = pa_schema.remove_metadata()
        return pa_schema


@_register_schema_internal
class ViewEmbedding(Embedding, ABC):
    """ViewEmbeddings are used to define an embedding vector for a view in a dataset.

    Attributes:
        view_ref: Reference to the embedding's view.
    """

    view_ref: ViewRef = ViewRef.none()

    @property
    def view(self) -> "View":
        """Get the embedding's view."""
        return self.resolve_ref(self.view_ref)

    @staticmethod
    def get_embedding_fn_from_table(dataset: "Dataset", table_name: str, metadata: dict) -> EmbeddingFunction:
        """Get the embedding function from a table.

        Args:
            dataset: The dataset containing the table.
            table_name: The name of the table containing the embedding function.
            metadata: The pyarrow metadata of the table.

        Returns:
            The embedding function.
        """
        registry = get_registry()

        serialized = metadata[b"embedding_functions"]
        raw_list = json.loads(serialized.decode("utf-8"))

        if len(raw_list) > 1:
            raise ValueError("Only one embedding function per table is supported")

        pixano_name = raw_list[0]["name"]
        if pixano_name not in registry._functions:
            name = _from_pixano_name(dataset, table_name, pixano_name)
            create_view_embedding_function(registry._functions[name], pixano_name, dataset)
        return registry.get(pixano_name)

    @classmethod
    def create_schema(
        cls,
        embedding_fn: str,
        table_name: str,
        dataset: "Dataset",
        **embedding_function_kwargs: Any,
    ) -> type["ViewEmbedding"]:
        """Create a ViewEmbedding schema.

        Args:
            embedding_fn: The embedding function.
            table_name: The name of the table containing the schema.
            dataset: The dataset to which the schema belongs.
            embedding_function_kwargs: The keyword arguments for creating the embedding function.

        Returns:
            The `ViewEmbedding` schema.
        """
        lance_registry = get_registry()
        if not isinstance(embedding_fn, str):
            raise TypeError(f"{embedding_fn} should be a string")

        pixano_name = _to_pixano_name(dataset, table_name, embedding_fn)
        if pixano_name not in lance_registry._functions:
            type_embedding_function = lance_registry.get(embedding_fn)
            view_embedding_function: type[EmbeddingFunction] = create_view_embedding_function(
                type_embedding_function, pixano_name, dataset
            )
        else:
            view_embedding_function = lance_registry.get(pixano_name)

        view_embedding_function = view_embedding_function.create(**embedding_function_kwargs)

        embedding_fields = {
            "vector": (Vector(view_embedding_function.ndims()), view_embedding_function.VectorField()),
            "view_ref": (ViewRef, view_embedding_function.SourceField()),
        }
        return create_model(
            "ViewEmbedding",
            __base__=ViewEmbedding,
            **embedding_fields,
        )


def is_embedding(cls: type, strict: bool = False) -> bool:
    """Check if a class is an `Embedding` or subclass of `Embedding`."""
    return issubclass_strict(cls, Embedding, strict)


def is_view_embedding(cls: type, strict: bool = False) -> bool:
    """Check if a class is an `ViewEmbedding` or subclass of `ViewEmbedding`."""
    return issubclass_strict(cls, ViewEmbedding, strict)


def create_view_embedding_function(
    type_embedding_function: type[EmbeddingFunction], name: str, dataset: "Dataset"
) -> type[EmbeddingFunction]:
    """Create a `ViewEmbeddingFunction` based on an
    [EmbeddingFunction][lancedb.embeddings.base.EmbeddingFunction].
    """

    @register(name)
    class ViewEmbeddingFunction(type_embedding_function):
        """A `ViewEmbeddingFunction` based on an [EmbeddingFunction][lancedb.embeddings.base.EmbeddingFunction]."""

        def _open_views(self, views: list[Any]) -> list[Any]:
            """Open the views in the dataset."""
            return [view.open(dataset.media_dir, "image") for view in views]

        def compute_source_embeddings(self, view_refs: pa.Table, *args, **kwargs) -> list:
            """Compute the embeddings for the source column in the database."""
            views = [dataset.resolve_ref(ViewRef(**view_ref)) for view_ref in view_refs.to_pylist()]
            view_type = type(views[0])
            if is_image(view_type) or is_sequence_frame(view_type):
                views = cast(list[Image], views)
                return super().compute_source_embeddings(self._open_views(views=views), *args, **kwargs)
            else:
                raise ValueError(f"View type {view_type} not supported for embedding.")

    return ViewEmbeddingFunction



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from .base import FolderBaseBuilder
from .image import ImageFolderBuilder
from .video import VideoFolderBuilder
from .vqa import VQAFolderBuilder


__all__ = ["ImageFolderBuilder", "FolderBaseBuilder", "VideoFolderBuilder", "VQAFolderBuilder"]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from .boxes import xywh_to_xyxy, xyxy_to_xywh
from .creators import create_instance_of_pixano_type, create_instance_of_schema
from .image import (
    binary_to_url,
    depth_array_to_gray,
    depth_file_to_binary,
    encode_rle,
    image_to_binary,
    mask_area,
    mask_to_polygons,
    mask_to_rle,
    polygons_to_rle,
    rle_to_mask,
    rle_to_polygons,
    rle_to_urle,
    urle_to_rle,
)


__all__ = [
    "binary_to_url",
    "create_instance_of_schema",
    "create_instance_of_pixano_type",
    "depth_array_to_gray",
    "depth_file_to_binary",
    "encode_rle",
    "image_to_binary",
    "mask_area",
    "mask_to_polygons",
    "mask_to_rle",
    "polygons_to_rle",
    "rle_to_urle",
    "rle_to_mask",
    "rle_to_polygons",
    "urle_to_rle",
    "xywh_to_xyxy",
    "xyxy_to_xywh",
]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pathlib import Path
from typing import Any, Iterator, Mapping

import pyarrow.json as pa_json
import shortuuid

from pixano.datasets.dataset_info import DatasetInfo
from pixano.datasets.dataset_schema import DatasetItem
from pixano.datasets.workspaces import WorkspaceType
from pixano.features import BaseSchema, Entity, Item, View, create_bbox, is_annotation, is_bbox, is_entity, is_view
from pixano.features.schemas.annotations.annotation import Annotation
from pixano.features.schemas.registry import _PIXANO_SCHEMA_REGISTRY
from pixano.features.schemas.source import SourceKind
from pixano.features.types.schema_reference import EntityRef, ItemRef, SourceRef, ViewRef
from pixano.features.utils.creators import create_instance_of_schema

from ..dataset_builder import DatasetBuilder


class FolderBaseBuilder(DatasetBuilder):
    """This is a class for building datasets based on a folder structure.

    The folder structure should be as follows:
        - source_dir/{split}/{item}.{ext}
        - source_dir/{split}/metadata.jsonl

    The metadata file should be a jsonl file with the following format:
    ```json
        [
            {
                "item": "item1",
                "metadata1": "value1",
                "metadata2": "value2",
                ...
                "entities": {
                    "attr1": [val1, val2, ...],
                    "attr2": [val1, val2, ...],
                    ...
                }
            },
            {
                "item": "item2",
                "metadata1": "value1",
                "metadata2": "value2",
                ...
                "entities": {
                    "attr1": [val1, val2, ...],
                    "attr2": [val1, val2, ...],
                    ...
                }
            },
            ...
        ]
    ```

    Note:
        Only one view and one entity are supported in folder builders.

    Attributes:
        source_dir: The source directory for the dataset.
        view_name: The name of the view schema.
        view_schema: The schema of the view.
        entity_name: The name of the entities schema.
        entity_schema: The schema of the entities.
        METADATA_FILENAME: The metadata filename.
        EXTENSIONS: The list of supported extensions.
        WORKSPACE_TYPE: The workspace type of the dataset. Subclass should override this attribute if workspace is
            known.
    """

    METADATA_FILENAME: str = "metadata.jsonl"
    EXTENSIONS: list[str]
    WORKSPACE_TYPE = WorkspaceType.UNDEFINED
    DEFAULT_SCHEMA: type[DatasetItem] | None = None

    def __init__(
        self,
        source_dir: Path | str,
        target_dir: Path | str,
        info: DatasetInfo,
        dataset_item: type[DatasetItem] | None = None,
        url_prefix: Path | str | None = None,
    ) -> None:
        """Initialize the `FolderBaseBuilder`.

        Args:
            source_dir: The source directory for the dataset.
            target_dir: The target directory for the dataset.
            dataset_item: The dataset item schema.
            info: User informations (name, description, ...) for the dataset.
            url_prefix: The path to build relative URLs for the views. Useful to build dataset libraries to pass the
                relative path from the media directory.
        """
        info.workspace = self.WORKSPACE_TYPE
        if self.DEFAULT_SCHEMA is not None and dataset_item is None:
            dataset_item = self.DEFAULT_SCHEMA
        if dataset_item is None:
            raise ValueError("A schema is required.")
        super().__init__(target_dir=target_dir, dataset_item=dataset_item, info=info)
        self.source_dir = Path(source_dir)
        if url_prefix is None:
            url_prefix = Path(".")
        else:
            url_prefix = Path(url_prefix)
        self.url_prefix = url_prefix

        self.views_schema: dict[str, type[View]] = {}
        self.entities_schema: dict[str, type[Entity]] = {}
        self.annotations_schema: dict[str, type[Annotation]] = {}

        for k, s in self.schemas.items():
            if is_view(s):
                self.views_schema.update({k: s})
            elif is_entity(s):
                self.entities_schema.update({k: s})
            elif is_annotation(s):
                self.annotations_schema.update({k: s})
        if not self.views_schema or not self.entities_schema:
            raise ValueError("At least one View and one Entity schema must be defined in the schemas argument.")

        # for compatibility with actual ImageFolderBuilder that allows only one view and one entity
        # TODO - allow multiview and multi entities in base FolderBuilder
        # Note: technically VQA also allow only one view, so for now we keep the ValueError
        if len(self.views_schema) == 1:
            self.view_name, self.view_schema = list(self.views_schema.items())[0]
        else:
            raise ValueError("Only one view schema is supported in folder based builders.")
        if len(self.entities_schema) == 1:
            self.entity_name, self.entity_schema = list(self.entities_schema.items())[0]

    def generate_data(
        self,
    ) -> Iterator[dict[str, BaseSchema | list[BaseSchema]]]:
        """Generate data from the source directory.

        Returns:
            An iterator over the data following the dataset schemas.
        """
        source_id = None
        for split in self.source_dir.glob("*"):
            if split.is_dir() and not split.name.startswith("."):
                try:
                    metadata = self._read_metadata(split / self.METADATA_FILENAME)
                except FileNotFoundError:
                    metadata = None
                if metadata is None:
                    for view_file in split.glob("*"):
                        # only consider {split}/{item}.{ext} files
                        if not view_file.is_file() or view_file.suffix not in self.EXTENSIONS:
                            continue
                        # create item
                        item = self._create_item(split.name, **{})
                        # create view
                        view = self._create_view(item, view_file, self.view_schema)
                        yield {
                            self.item_schema_name: item,
                            self.view_name: view,
                        }
                else:  # metadata not None
                    # retrieve item metadata in metadata file
                    for item_metadata in metadata:
                        # extract entity metadata from item metadata
                        entities_data = item_metadata.pop(self.entity_name, None)

                        # create item
                        item = self._create_item(split.name, **item_metadata)

                        # create view
                        view_file = split / item_metadata[self.view_name]
                        view = self._create_view(item, view_file, self.view_schema)

                        if entities_data is None:
                            yield {
                                self.item_schema_name: item,
                                self.view_name: view,
                            }
                            continue
                        elif source_id is None:
                            source_id = self.add_source("Builder", SourceKind.OTHER)

                        # create entities and their annotations
                        entities, annotations = self._create_entities(item, view, entities_data, source_id)

                        yield {
                            self.item_schema_name: item,
                            self.view_name: view,
                            self.entity_name: entities,
                            **annotations,
                        }

    def _create_item(self, split: str, **item_metadata) -> BaseSchema:
        if "id" not in item_metadata:
            item_metadata["id"] = shortuuid.uuid()
        return self.item_schema(
            split=split,
            **item_metadata,
        )

    def _create_view(self, item: Item, view_file: Path, view_schema: type[View]) -> View:
        if not issubclass(view_schema, View):
            raise ValueError("View schema must be a subclass of View")
        if view_schema not in _PIXANO_SCHEMA_REGISTRY.values():
            raise ValueError(
                f"View schema {view_schema} is not supported. You should implement your own _create_view method."
            )

        view = create_instance_of_schema(
            view_schema,
            id=shortuuid.uuid(),
            item_ref=ItemRef(id=item.id),
            url=view_file,
            url_relative_path=self.source_dir,
        )
        view.url = str(self.url_prefix / view.url)
        return view

    def _create_entities(
        self, item: Item, view: View, entities_data: dict[str, Any], source_id: str
    ) -> tuple[list[Entity], dict[str, list[Annotation]]]:
        entities: list[Entity] = []
        annotations: dict[str, list[Annotation]] = {}

        entities_attrs = list(entities_data.keys())
        for attr in entities_attrs:
            if attr not in self.entity_schema.model_fields.keys() and attr not in self.schemas.keys():
                raise ValueError(f"Attribute {attr} not found in entity schema and is not a schema annotation.")

        # check if all list of entities data have same length
        nums_entities = {len(v) if isinstance(v, list) else 1 for v in entities_data.values() if v is not None}
        if len(nums_entities) > 1:
            raise ValueError("All list of entities data must have same length")
        elif len(nums_entities) == 0:
            return [], {}

        num_entities = nums_entities.pop()
        entities_data = {k: v if isinstance(v, list) else [v] for k, v in entities_data.items() if v is not None}
        entities_attrs = list(entities_data.keys())

        for i in range(num_entities):
            entity = {}
            entity_annotations: dict[str, Any] = {}
            entity_id = shortuuid.uuid()
            for attr in entities_attrs:
                if entities_data[attr] is None:
                    continue

                # check if attribute is an annotation schema
                is_attr_schema = attr in self.schemas
                if is_attr_schema and not issubclass(self.schemas[attr], Annotation):
                    raise ValueError(f"Attribute {attr} must be a subclass of Annotation")

                # create annotation if attribute is an annotation schema
                if is_attr_schema:
                    if attr not in entity_annotations:
                        entity_annotations[attr] = []

                    schema = self.schemas[attr]
                    if isinstance(entities_data[attr][i], Mapping):
                        annotation = create_instance_of_schema(
                            schema,
                            id=shortuuid.uuid(),
                            item_ref=ItemRef(id=item.id),
                            view_ref=ViewRef(id=view.id, name=self.view_name),
                            entity_ref=EntityRef(id=entity_id, name=self.entity_name),
                            source_ref=SourceRef(id=source_id),
                            **entities_data[attr][i],
                        )
                    else:
                        # TODO check jsonl format for mask & keypoints
                        if is_bbox(schema, True):
                            annotation = create_bbox(
                                id=shortuuid.uuid(),
                                item_ref=ItemRef(id=item.id),
                                view_ref=ViewRef(id=view.id, name=self.view_name),
                                entity_ref=EntityRef(id=entity_id, name=self.entity_name),
                                source_ref=SourceRef(id=source_id),
                                coords=entities_data[attr][i],
                                format="xywh",
                                is_normalized=True,
                                confidence=1.0,
                            )
                        else:
                            raise ValueError(f"Schema {schema} not supported for infered entity creation.")
                    entity_annotations[attr].append(annotation)
                else:
                    entity[attr] = entities_data[attr][i]
            entities.append(
                self.entity_schema(
                    id=entity_id,
                    item_ref=ItemRef(id=item.id),
                    view_ref=ViewRef(id=view.id, name=self.view_name),
                    **entity,
                )
            )

            for key, entity_annotation in entity_annotations.items():
                if key not in annotations:
                    annotations[key] = []
                annotations[key].extend(entity_annotation)
        return entities, annotations

    def _read_metadata(self, metadata_file: Path) -> list[dict]:
        if not metadata_file.exists():
            raise FileNotFoundError(f"Metadata file {metadata_file} not found")
        return pa_json.read_json(metadata_file).to_pylist()



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import io
from pathlib import Path
from typing import Literal, overload
from urllib.error import URLError
from urllib.parse import urlparse
from urllib.request import urlopen

import PIL.Image
from PIL.Image import Image as PILImage

from pixano.features.utils.image import image_to_base64
from pixano.utils import issubclass_strict

from ...types.schema_reference import ItemRef, ViewRef
from ..registry import _register_schema_internal
from .view import View


@_register_schema_internal
class Image(View):
    """Image view.

    Attributes:
        url: The image URL. Can be relative or absolute or a data URL.
        width: The image width.
        height: The image height.
        format: The image format.
    """

    url: str
    width: int
    height: int
    format: str

    @overload
    def open(self, media_dir: Path | None, output_type: Literal["base64"] = "base64") -> str: ...
    @overload
    def open(self, media_dir: Path | None, output_type: Literal["image"]) -> PILImage: ...
    def open(
        self,
        media_dir: Path | None = None,
        output_type: Literal["base64", "image"] = "base64",
    ) -> str | PILImage:
        """Open the image.

        Note:
            If the output type is "base64", the image is returned as a base64 string formatted as
            "data:image/{image_format};base64,{base64}".

        Args:
            media_dir: Path to the media directory. If the URL is relative, it is relative to this directory.
            output_type: The output type. Can be "base64" or "image" (PIL.Image).

        Returns:
            opened image.
        """
        return Image.open_url(url=self.url, media_dir=media_dir, output_type=output_type)

    @overload
    @staticmethod
    def open_url(
        url: str,
        media_dir: Path | None,
        output_type: Literal["base64"] = "base64",
    ) -> str: ...
    @overload
    @staticmethod
    def open_url(
        url: str,
        media_dir: Path | None,
        output_type: Literal["image"],
    ) -> PILImage: ...
    @staticmethod
    def open_url(
        url: str,
        media_dir: Path | None = None,
        output_type: Literal["base64", "image"] = "base64",
    ) -> str | PILImage:
        """Open an image from a URL.

        Note:
            If the output type is "base64", the image is returned as a base64 string formatted as
            "data:image/{image_format};base64,{base64}".

        Args:
            url: image url relative to media_dir or absolute.
            media_dir: path to the media directory if the URL is relative.
            output_type: output type. Can be "base64" or "image" (PIL.Image).

        Returns:
            The opened image.
        """
        if output_type not in ["base64", "image"]:
            raise ValueError(f"Invalid output type: {output_type}")

        # URI is incomplete
        if urlparse(url).scheme == "":
            if media_dir is None:
                raise ValueError("URI is incomplete, need media directory")
            uri_prefix = media_dir.absolute().as_uri()
            # URI prefix exists
            if uri_prefix is not None:
                parsed_uri = urlparse(uri_prefix)
                # URI prefix is incomplete
                if parsed_uri.scheme == "":
                    raise ValueError("URI prefix is incomplete, no scheme provided (http://, file://, ...)")
                if url.startswith("/"):
                    url = url[1:]
                combined_path = Path(parsed_uri.path) / url
                parsed_uri = parsed_uri._replace(path=str(combined_path))
                api_url = parsed_uri.geturl()
            else:
                # No URI prefix
                raise ValueError("URI is incomplete, need URI prefix")
        # URI is already complete
        else:
            api_url = url

        try:
            with urlopen(api_url) as f:
                im_bytes = f.read()
        except URLError:
            raise ValueError(f"Error: image not found ({api_url})")

        pil_image = PIL.Image.open(io.BytesIO(im_bytes))

        # Handle output types
        if output_type == "base64":
            return image_to_base64(pil_image)

        return pil_image

    def shorten_url_to_relative_path(self, url_relative_path: Path) -> str:
        """Changes the URL of an image to be relative.

        Note:
           This helps the creation of a dataset where Image object are created
           with known dimensions format and theorical path, but images are not yet accessible.

        Args:
            url_relative_path (Path): The path to convert the URL to a relative path,
              eg for images to be later searchable in the media_dir.

        Returns:
            str: shorten image url
        """
        url = Path(self.url)
        url_relative_path = Path(url_relative_path)
        self.url = url.relative_to(url_relative_path).as_posix()
        return self.url


def is_image(cls: type, strict: bool = False) -> bool:
    """Check if the given class is `Image` or a subclass of `Image`."""
    return issubclass_strict(cls, Image, strict)


def create_image(
    url: Path,
    id: str = "",
    item_ref: ItemRef = ItemRef.none(),
    parent_ref: ViewRef = ViewRef.none(),
    width: int | None = None,
    height: int | None = None,
    format: str | None = None,
    url_relative_path: Path | None = None,
) -> Image:
    """Create an `Image` instance.

    Args:
        url: The image URL. If not relative, the URL is converted to a relative path using `url_relative_path`.
        id: Image ID.
        item_ref: Item reference.
        parent_ref: Parent view reference.
        width: The image width. If None, the width is extracted from the image file.
        height: The image height. If None, the height is extracted from the image file.
        format: The image format. If None, the format is extracted from the image file.
        url_relative_path: The path to convert the URL to a relative path,
           eg for images to be searchable in the media_dir.

    Returns:
        The created `Image` instance.
    """
    none_conditions = [width is None, height is None, format is None]
    not_none_conditions = [width is not None, height is not None, format is not None]
    if not all(none_conditions) and not all(not_none_conditions):
        raise ValueError("width, height and format must be all defined or all None")

    url = Path(url)

    if width is None:
        img = PIL.Image.open(url)
        width = img.width
        height = img.height
        format = img.format

    if url_relative_path is not None:
        url_relative_path = Path(url_relative_path)
        url = url.relative_to(url_relative_path)

    return Image(
        id=id,
        item_ref=item_ref,
        parent_ref=parent_ref,
        url=str(url.as_posix()),
        width=width,
        height=height,
        format=format,
    )



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pathlib import Path
from typing import Any

import shortuuid
from fastapi.encoders import jsonable_encoder
from pixano_inference.client import PixanoInferenceClient
from pixano_inference.pydantic import LanceVector
from pixano_inference.pydantic.tasks import CompressedRLE as PixanoInferenceCompressedRLE
from pixano_inference.pydantic.tasks import ImageMaskGenerationRequest, ImageMaskGenerationResponse
from pixano_inference.utils import is_url

from pixano.features import BBox, CompressedRLE, Image, NDArrayFloat, Source, ViewEmbedding
from pixano.features.schemas.entities.entity import Entity
from pixano.features.types.schema_reference import EntityRef, SourceRef, ViewRef


async def image_mask_generation(
    client: PixanoInferenceClient,
    media_dir: Path,
    image: Image,
    entity: Entity,
    source: Source,
    image_embedding: ViewEmbedding | NDArrayFloat | LanceVector | None = None,
    high_resolution_features: list[ViewEmbedding] | list[NDArrayFloat] | list[LanceVector] | None = None,
    bbox: BBox | None = None,
    points: list[list[int]] | None = None,
    labels: list[int] | None = None,
    **client_kwargs: Any,
) -> tuple[CompressedRLE, float, NDArrayFloat | None, list[NDArrayFloat] | None]:
    """Image mask generation task.

    Args:
        client: Pixano inference client.
        media_dir: Media directory.
        image: Image to generate mask for.
        entity: Entity associated with the image.
        source: The source refering to the model.
        image_embedding: Image embedding.
        high_resolution_features: High resolution features.
        bbox: Bounding box of the object in the original image.
        points: Points to generate mask for.
        labels: Labels of the points. If 0, the point is background else the point is foreground.
        client_kwargs: Additional kwargs for the client to be passed.

    Returns:
        tuple of the compressed RLE mask, its score, the source of the image, the image embeddings and the high
        resolution features The features are returned if not provided in the arguments otherwise None is returned.
    """
    image_request = image.url if is_url(image.url) else image.open(media_dir, "base64")
    if image_embedding is None or isinstance(image_embedding, (LanceVector, NDArrayFloat)):
        image_embedding_request = image_embedding
    else:
        image_embedding_request = NDArrayFloat(values=image_embedding.vector, shape=image_embedding.shape)
    if high_resolution_features is None:
        high_resolution_features_request = None
    else:
        high_resolution_features_request = []
        for feature in high_resolution_features:
            if isinstance(feature, (LanceVector, NDArrayFloat)):
                high_resolution_features_request.append(feature)
            else:
                high_resolution_features_request.append(NDArrayFloat(values=feature.vector, shape=feature.shape))

    if points is not None:
        points_request = [points]
    else:
        points_request = None
    if labels is not None:
        labels_request = [labels]
    else:
        labels_request = None
    if bbox is not None:
        if bbox.is_normalized:
            bbox = bbox.denormalize(height=image.height, width=image.width)
        bbox_request = [[int(c) for c in bbox.xyxy_coords]]
    else:
        bbox_request = None

    return_image_embedding = image_embedding is None

    request = ImageMaskGenerationRequest(
        image=image_request,
        image_embedding=image_embedding_request,
        high_resolution_features=high_resolution_features_request,
        boxes=bbox_request,
        points=points_request,
        labels=labels_request,
        num_multimask_outputs=1,
        multimask_output=False,
        return_image_embedding=return_image_embedding,
        model=source.name,
    )

    response: ImageMaskGenerationResponse = await client.image_mask_generation(request, **client_kwargs)

    inference_metadata = jsonable_encoder(
        {
            "timestamp": response.timestamp,
            "processing_time": response.processing_time,
            **response.metadata,
        }
    )

    mask_inference: PixanoInferenceCompressedRLE = response.data.masks[0][0]
    mask = CompressedRLE(
        id=shortuuid.uuid(),
        item_ref=image.item_ref,
        view_ref=ViewRef(name=image.table_name, id=image.id),
        entity_ref=EntityRef(name=entity.table_name, id=entity.id),
        source_ref=SourceRef(id=source.id),
        inference_metadata=inference_metadata,
        **mask_inference.model_dump(),
    )
    score = response.data.scores.values[0]

    return mask, score, response.data.image_embedding, response.data.high_resolution_features



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================


from pathlib import Path

from pixano.utils import issubclass_strict

from ...types.schema_reference import ItemRef, ViewRef
from ..registry import _register_schema_internal
from .image import Image, create_image


@_register_schema_internal
class SequenceFrame(Image):
    """Sequence Frame view.

    Attributes:
        timestamp: The timestamp of the frame.
        frame_index: The index of the frame in the sequence.
    """

    timestamp: float
    frame_index: int


def is_sequence_frame(cls: type, strict: bool = False) -> bool:
    """Check if the given class is a subclass of `SequenceFrame`."""
    return issubclass_strict(cls, SequenceFrame, strict)


def create_sequence_frame(
    url: Path,
    timestamp: float,
    frame_index: int,
    id: str = "",
    item_ref: ItemRef = ItemRef.none(),
    parent_ref: ViewRef = ViewRef.none(),
    width: int | None = None,
    height: int | None = None,
    format: str | None = None,
    url_relative_path: Path | None = None,
) -> SequenceFrame:
    """Create a `SequenceFrame` instance.

    Args:
        url: The frame URL. If not relative, the URL is converted to a relative path using `other_path`.
        timestamp: The timestamp of the frame.
        frame_index: The index of the frame in the sequence.
        id: Point cloud ID.
        item_ref: Item reference.
        parent_ref: Parent view reference.
        width: The frame width. If None, the width is extracted from the frame file.
        height: The frame height. If None, the height is extracted from the frame file.
        format: The frame format. If None, the format is extracted from the frame file.
        url_relative_path: The path to convert the URL to a relative path.

    Returns:
        The created `SequenceFrame` instance.
    """
    image = create_image(url, id, item_ref, parent_ref, width, height, format, url_relative_path)
    return SequenceFrame(
        id=id,
        item_ref=item_ref,
        parent_ref=parent_ref,
        url=image.url,
        width=image.width,
        height=image.height,
        format=image.format,
        timestamp=timestamp,
        frame_index=frame_index,
    )



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from .image import Image, create_image, is_image
from .point_cloud import PointCloud, create_point_cloud, is_point_cloud
from .sequence_frame import SequenceFrame, create_sequence_frame, is_sequence_frame
from .text import Text, create_text, is_text
from .video import Video, create_video, is_video
from .view import View, is_view


__all__ = [
    "Image",
    "View",
    "Video",
    "SequenceFrame",
    "PointCloud",
    "Text",
    "create_image",
    "create_point_cloud",
    "create_sequence_frame",
    "create_text",
    "create_video",
    "is_image",
    "is_point_cloud",
    "is_sequence_frame",
    "is_view",
    "is_video",
    "is_text",
]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from typing_extensions import TYPE_CHECKING


if TYPE_CHECKING:
    from pixano.features.schemas import BaseSchema
    from pixano.features.types import BaseType


def create_instance_of_schema(schema: type["BaseSchema"], **data) -> "BaseSchema":
    """Create a row in a Schema."""
    # Import here to avoid circular imports
    from pixano.features.schemas import (
        create_bbox,
        create_bbox3d,
        create_cam_calibration,
        create_compressed_rle,
        create_image,
        create_keypoints,
        create_keypoints3d,
        create_sequence_frame,
        create_track,
        create_tracklet,
        create_video,
        is_base_schema,
        is_bbox,
        is_bbox3d,
        is_cam_calibration,
        is_compressed_rle,
        is_image,
        is_keypoints,
        is_keypoints3d,
        is_sequence_frame,
        is_track,
        is_tracklet,
        is_video,
    )

    if is_image(schema, strict=True):
        return create_image(**data)

    elif is_video(schema, strict=True):
        return create_video(**data)

    elif is_sequence_frame(schema, strict=True):
        return create_sequence_frame(**data)

    elif is_tracklet(schema, strict=True):
        return create_tracklet(**data)

    elif is_track(schema, True):
        return create_track(**data)

    elif is_bbox(schema, True):
        return create_bbox(**data)

    elif is_bbox3d(schema, True):
        return create_bbox3d(**data)

    elif is_cam_calibration(schema, True):
        return create_cam_calibration(**data)

    elif is_compressed_rle(schema, True):
        return create_compressed_rle(**data)

    elif is_keypoints(schema, True):
        return create_keypoints(**data)

    elif is_keypoints3d(schema, True):
        return create_keypoints3d(**data)

    elif is_base_schema(schema, False):
        return schema(**data)

    raise ValueError(f"Schema {schema} is not a base schema.")


def create_instance_of_pixano_type(pix_type: type["BaseType"], **data) -> "BaseType":
    """Create a pixano object."""
    # Import here to avoid circular imports
    from pixano.features.types import (
        create_annotation_ref,
        create_embedding_ref,
        create_entity_ref,
        create_item_ref,
        create_ndarray_float,
        create_schema_ref,
        create_source_ref,
        create_view_ref,
        is_annotation_ref,
        is_base_type,
        is_embedding_ref,
        is_entity_ref,
        is_item_ref,
        is_ndarray_float,
        is_schema_ref,
        is_source_ref,
        is_view_ref,
    )

    if is_ndarray_float(pix_type, True):
        return create_ndarray_float(**data)

    elif is_schema_ref(pix_type, True):
        return create_schema_ref(**data)

    elif is_item_ref(pix_type, True):
        return create_item_ref(**data)

    elif is_view_ref(pix_type, True):
        return create_view_ref(**data)

    elif is_entity_ref(pix_type, True):
        return create_entity_ref(**data)

    elif is_annotation_ref(pix_type, True):
        return create_annotation_ref(**data)

    elif is_embedding_ref(pix_type, True):
        return create_embedding_ref(**data)

    elif is_source_ref(pix_type, True):
        return create_source_ref(**data)

    elif is_base_type(pix_type, False):
        return pix_type(**data)

    raise ValueError(f"Type {pix_type} not supported.")



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import asyncio
import warnings
from functools import lru_cache

import click
import fastapi
import pkg_resources  # type: ignore[import-untyped]
import uvicorn
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from pixano_inference.client import PixanoInferenceClient

from pixano.app.display import display_cli, display_colab, display_ipython
from pixano.app.main import create_app
from pixano.app.settings import Settings, get_settings


LOGO = """
                             ▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▒
                           ▒▓▒                                       ▓
              █▓          ▓▓                                         ▓
                        ▒▓▒   ▒▓▓▓▓▓ ▒▓   ▓▒▒▓▓▓▓▒     ▒▓▓▓▓▓▒       ▓
▓▓ ▓▓▓▓▓▓▒    ▓▒  ▒▓   ▓▒   ▒▓      ▓██   ██     ▓▒  ▒█▓     ▒█▓     ▓
███      ▓▓   █▒   ▓█      ▒█▒       ██   █▒     ▒█  ▓▒        █     ▓
██        █▒  █▒    ▒█▓     █▒       ██   █▒     ▒█  ▓▓        █     ▓
██        █▒  █▒    ▓█▓█     █▒     ▓██   █▒     ▒█   ██     ▓█▒     ▓
██▓▒    ▒▓▒   █▒   ▓▒  ▒█▓    ▒▓▓▓▓▓ ▒▓   ▓      ▒▓    ▒▓▓▓▓▓▒       ▓
██  ▒▓▓▒▒     ▒   ▒      ▒▓▒                                         ▓
██                         █▓                                        ▓
"""
ASSETS_PATH = pkg_resources.resource_filename("pixano", "app/dist/_app")
TEMPLATE_PATH = pkg_resources.resource_filename("pixano", "app/dist")

task_functions = {
    "colab": asyncio.get_event_loop().create_task,
    "ipython": asyncio.get_event_loop().create_task,
    "none": asyncio.run,
}
display_functions = {
    "colab": display_colab,
    "ipython": display_ipython,
    "none": display_cli,
}


class App:
    """The Pixano app.

    Attributes:
        app: FastAPI App.
        config: App config.
        server: App server.
    """

    app: fastapi.FastAPI
    config: uvicorn.Config
    server: uvicorn.Server

    def __init__(
        self,
        library_dir: str,
        media_dir: str,
        models_dir: str | None = None,
        aws_endpoint: str | None = None,
        aws_region: str | None = None,
        aws_access_key: str | None = None,
        aws_secret_key: str | None = None,
        host: str = "127.0.0.1",
        port: int = 8000,
        pixano_inference_url: str | None = None,
    ):
        """Initialize and serve the Pixano app.

        Args:
            library_dir: Local or S3 path to dataset library
            media_dir: Local or S3 path to media library
            models_dir: Path to your models. If not provided, and
                library_dir is not a S3 path, it is set to
                library_dir/models.
            aws_endpoint: S3 endpoint URL, use 'AWS' if not provided.
                Used if library_dir is an S3 path.
            aws_region: S3 region name, not always required for
                private storages. Used if library_dir is an S3 path.
            aws_access_key: S3 AWS access key. Used if library_dir is
                an S3 path.
            aws_secret_key: S3 AWS secret key. Used if library_dir
                is an S3 path.
            host: App host.
            port: App port.
            pixano_inference_url: Pixano inference URL if any.
        """

        # Override app settings
        @lru_cache
        def get_settings_override():
            return Settings(
                library_dir=library_dir,
                media_dir=media_dir,
                models_dir=models_dir,
                aws_endpoint=aws_endpoint,
                aws_region=aws_region,
                aws_access_key=aws_access_key,
                aws_secret_key=aws_secret_key,
                pixano_inference_client=PixanoInferenceClient.connect(pixano_inference_url)
                if pixano_inference_url is not None
                else None,
            )

        # Create app
        settings = get_settings_override()
        templates = Jinja2Templates(directory=TEMPLATE_PATH)
        self.app = create_app(settings=settings)
        self.app.dependency_overrides[get_settings] = get_settings_override

        @self.app.get("/", response_class=HTMLResponse)
        def main_page(request: fastapi.Request):
            return templates.TemplateResponse("index.html", {"request": request})

        @self.app.get("/{ds_id}/dataset", response_class=HTMLResponse)
        async def dataset_page(request: fastapi.Request):
            return templates.TemplateResponse("index.html", {"request": request})

        @self.app.get("/{ds_id}/dashboard", response_class=HTMLResponse)
        async def dashboard_page(request: fastapi.Request):
            return templates.TemplateResponse("index.html", {"request": request})

        @self.app.get("/{ds_id}/dataset/{item_id}", response_class=HTMLResponse)
        async def item_page(request: fastapi.Request):
            return templates.TemplateResponse("index.html", {"request": request})

        try:
            self.app.mount("/_app", StaticFiles(directory=ASSETS_PATH), name="assets")
        # TODO: properly define environment variable for production to raise a RuntimeError accordingly
        except RuntimeError:
            warnings.warn(
                "Pixano app assets not found. If it is a production environment, this is not expected, "
                "check if you have built the assets for the UI."
            )
        self.config = uvicorn.Config(self.app, host=host, port=port)
        self.server = uvicorn.Server(self.config)

        # Serve app
        task_functions[self.get_env()](self.server.serve())  # type: ignore[operator]

    def display(self, height: int = 1000) -> None:
        """Display the Pixano app.

        Args:
            height: Frame height.
        """
        # Wait for app to be online
        while not self.server.started:
            task_functions[self.get_env()](asyncio.wait(0.1))  # type: ignore[operator, call-overload]

        # Display app
        for server in self.server.servers:
            for socket in server.sockets:
                address = socket.getsockname()
                display_functions[self.get_env()](url=f"http://{address[0]}", port=address[1], height=height)  # type: ignore[operator]

    def get_env(self) -> str:
        """Get the app's current running environment.

        Returns:
            Running environment.
        """
        # If Google colab import succeeds
        try:
            import google.colab  # noqa: F401, I001 #type: ignore
            import IPython
        except ImportError:
            pass
        else:
            if IPython.get_ipython() is not None:
                return "colab"

        # If IPython import succeeds
        try:
            import IPython
        except ImportError:
            pass
        else:
            ipython = IPython.get_ipython()
            if ipython is not None and ipython.has_trait("kernel"):
                return "ipython"

        # Else
        return "none"


@click.command(context_settings={"auto_envvar_prefix": "UVICORN"})
@click.argument("library_dir", type=str)
@click.argument("media_dir", type=str)
@click.option(
    "--models_dir",
    type=str,
    help="Path to your models. If not provided, and library_dir is not a S3 path, it is set to library_dir/models",
)
@click.option(
    "--aws_endpoint",
    type=str,
    help=("S3 endpoint URL, use 'AWS' if not provided. Used if library_dir or media_dir is an S3 path"),
)
@click.option(
    "--aws_region",
    type=str,
    help=("S3 region name, not always required for private storages. Used if library_dir or media_dir is an S3 path"),
)
@click.option(
    "--aws_access_key",
    type=str,
    help="S3 AWS access key. Used if library_dir or media_dir is an S3 path",
)
@click.option(
    "--aws_secret_key",
    type=str,
    help="S3 AWS secret key. Used if library_dir or media_dir is an S3 path",
)
@click.option(
    "--host",
    type=str,
    default="127.0.0.1",
    help="Pixano app URL host",
    show_default=True,
)
@click.option(
    "--port",
    type=int,
    default=0,
    help="Pixano app URL port",
    show_default=True,
)
@click.option("--pixano_inference_url", type=str, help="Pixano inference API url.", default="")
def main(
    library_dir: str,
    media_dir: str,
    models_dir: str,
    aws_endpoint: str,
    aws_region: str,
    aws_access_key: str,
    aws_secret_key: str,
    host: str,
    port: int,
    pixano_inference_url: str,
) -> None:
    """Launch Pixano App in LIBRARY_DIR.

    Args:
        library_dir: Local or S3 path to dataset library.
        media_dir: Local or S3 path to media library.
        models_dir: Path to your models. If not provided, and
            library_dir is not a S3 path, it is set to
            library_dir/models.
        aws_endpoint: S3 endpoint URL, use 'AWS' if not provided.
        aws_region: S3 region name, not always required for private storages.
        aws_access_key: S3 AWS access key.
        aws_secret_key: S3 AWS secret key.
        host: App host.
        port: App port.
        pixano_inference_url: Pixano inference URL.
    """
    App(
        library_dir,
        media_dir,
        models_dir,
        aws_endpoint,
        aws_region,
        aws_access_key,
        aws_secret_key,
        host,
        port,
        pixano_inference_url if pixano_inference_url != "" else None,
    )


if __name__ == "__main__":
    main()



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pathlib import Path

import lancedb
from pydantic import ConfigDict

from pixano.datasets import DatasetInfo
from pixano.datasets.dataset import Dataset
from pixano.features.schemas.schema_group import SchemaGroup


class DatasetInfoModel(DatasetInfo):
    """Dataset info model.

    It contains all the information as a [DatasetInfo][pixano.datasets.dataset_info.DatasetInfo] and the number of
    items in the dataset.

    Attributes:
        num_items: Number of items in the dataset.
    """

    model_config = ConfigDict(
        validate_assignment=True,
        json_schema_extra={
            "examples": [
                {
                    "id": "dataset_id",
                    "name": "Dataset name",
                    "description": "Dataset description",
                    "size": "Unknown",
                    "preview": "path/to/preview.jpg",
                    "num_items": 100,
                    "workspace": "image",
                }
            ]
        },
    )

    num_items: int

    @staticmethod
    def from_dataset_info(dataset_info: DatasetInfo, dataset_dir: Path) -> "DatasetInfoModel":
        """Create a dataset info model from a dataset info.

        Args:
            dataset_info: Dataset info.
            dataset_dir: Dataset directory.

        Returns:
            Dataset info model.
        """
        db = lancedb.connect(dataset_dir / Dataset._DB_PATH)
        item_table = db.open_table(SchemaGroup.ITEM.value)
        num_items = item_table.count_rows()
        return DatasetInfoModel(
            num_items=num_items,
            **dataset_info.model_dump(),
        )



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from typing import Annotated

from fastapi import APIRouter, Depends, Query

from pixano.app.models.entities import EntityModel
from pixano.app.settings import Settings, get_settings
from pixano.features.schemas.schema_group import SchemaGroup

from .utils import (
    create_row_handler,
    create_rows_handler,
    delete_row_handler,
    delete_rows_handler,
    get_row_handler,
    get_rows_handler,
    update_row_handler,
    update_rows_handler,
)


router = APIRouter(prefix="/entities", tags=["Entitys"])


@router.get("/{dataset_id}/{table}/", response_model=list[EntityModel])
async def get_entities(
    dataset_id: str,
    table: str,
    settings: Annotated[Settings, Depends(get_settings)],
    ids: list[str] | None = Query(None),
    limit: int | None = None,
    skip: int = 0,
    where: str | None = None,
    item_ids: list[str] | None = Query(None),
) -> list[EntityModel]:
    """Get entities from a table of a dataset.

    They can be filtered by IDs, item IDs, a where clause or paginated.

    Args:
        dataset_id: Dataset ID containing the table.
        table: Table name.
        settings: App settings.
        ids: IDs of the views.
        limit: Limit number of views.
        skip: Skip number of views.
        where: Where clause.
        item_ids: Item IDs.

    Returns:
        List of views.
    """
    return await get_rows_handler(
        dataset_id=dataset_id,
        group=SchemaGroup.ENTITY,
        table=table,
        settings=settings,
        where=where,
        ids=ids,
        item_ids=item_ids,
        limit=limit,
        skip=skip,
    )


@router.get("/{dataset_id}/{table}/{id}", response_model=EntityModel)
async def get_entity(
    dataset_id: str, table: str, id: str, settings: Annotated[Settings, Depends(get_settings)]
) -> EntityModel:
    """Get an entity from a table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        table: Table name.
        id: ID of the entity.
        settings: App settings.

    Returns:
        The entity.
    """
    return await get_row_handler(dataset_id, SchemaGroup.ENTITY, table, id, settings)


@router.post("/{dataset_id}/{table}/", response_model=list[EntityModel])
async def create_entities(
    dataset_id: str,
    table: str,
    entities: list[EntityModel],
    settings: Annotated[Settings, Depends(get_settings)],
) -> list[EntityModel]:
    """Add entities in a table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        table: Table name.
        entities: Entities to add.
        settings: App settings.

    Returns:
        List of entities added.
    """
    return await create_rows_handler(dataset_id, SchemaGroup.ENTITY, table, entities, settings)


@router.post("/{dataset_id}/{table}/{id}", response_model=EntityModel)
async def create_entity(
    dataset_id: str,
    table: str,
    id: str,
    entity: EntityModel,
    settings: Annotated[Settings, Depends(get_settings)],
) -> EntityModel:
    """Add an entity in a table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        table: Table name.
        id: ID of the entity.
        entity: Entity to add.
        settings: App settings.

    Returns:
        The entity added.
    """
    return await create_row_handler(dataset_id, SchemaGroup.ENTITY, table, id, entity, settings)


@router.put("/{dataset_id}/{table}/{id}", response_model=EntityModel)
async def update_entity(
    dataset_id: str,
    table: str,
    id: str,
    entity: EntityModel,
    settings: Annotated[Settings, Depends(get_settings)],
) -> EntityModel:
    """Update an entity in a table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        table: Table name.
        id: ID of the entity.
        entity: Entity to update.
        settings: App settings.

    Returns:
        The entity updated.
    """
    return await update_row_handler(dataset_id, SchemaGroup.ENTITY, table, id, entity, settings)


@router.put("/{dataset_id}/{table}/", response_model=list[EntityModel])
async def update_entities(
    dataset_id: str,
    table: str,
    entities: list[EntityModel],
    settings: Annotated[Settings, Depends(get_settings)],
) -> list[EntityModel]:
    """Update entities in a table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        table: Table name.
        entities: Entities to update.
        settings: App settings.

    Returns:
        List of entities updated.
    """
    return await update_rows_handler(dataset_id, SchemaGroup.ENTITY, table, entities, settings)


@router.delete("/{dataset_id}/{table}/{id}")
async def delete_entity(
    dataset_id: str, table: str, id: str, settings: Annotated[Settings, Depends(get_settings)]
) -> None:
    """Delete an entity from a table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        table: Table name.
        id: ID of the entity to delete.
        settings: App settings.
    """
    return await delete_row_handler(dataset_id, SchemaGroup.ENTITY, table, id, settings)


@router.delete("/{dataset_id}/{table}/")
async def delete_entities(
    dataset_id: str,
    table: str,
    ids: Annotated[list[str], Query()],
    settings: Annotated[Settings, Depends(get_settings)],
) -> None:
    """Delete entities from a table of a dataset.

    Args:
        dataset_id: Dataset ID containing the table.
        table: Table name.
        ids: IDs of the entities to delete.
        settings: App settings.
    """
    return await delete_rows_handler(dataset_id, SchemaGroup.ENTITY, table, ids, settings)



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

import re
from collections import defaultdict
from datetime import datetime
from pathlib import Path
from typing import Any, Iterator

import pyarrow.json as pa_json
import shortuuid

from pixano.datasets.dataset_schema import DatasetItem
from pixano.datasets.utils import mosaic
from pixano.datasets.workspaces import DefaultVQADatasetItem, WorkspaceType
from pixano.features import BaseSchema, Entity, Item, Message, View, is_conversation
from pixano.features.schemas.annotations import Annotation
from pixano.features.schemas.registry import _PIXANO_SCHEMA_REGISTRY
from pixano.features.schemas.source import SourceKind
from pixano.features.types.schema_reference import EntityRef, ItemRef, SourceRef, ViewRef
from pixano.features.utils.creators import create_instance_of_schema

from .image import ImageFolderBuilder


class VQAFolderBuilder(ImageFolderBuilder):
    """Builder for vqa datasets stored in a folder."""

    WORKSPACE_TYPE = WorkspaceType.IMAGE_VQA
    DEFAULT_SCHEMA: type[DatasetItem] = DefaultVQADatasetItem

    def generate_data(
        self,
    ) -> Iterator[dict[str, BaseSchema | list[BaseSchema]]]:
        """Generate data from the source directory.

        Returns:
            An iterator over the data following the dataset schemas.
        """
        source_id = None
        for split in self.source_dir.glob("*"):
            if split.is_dir() and not split.name.startswith("."):
                try:
                    dataset_pieces = self._read_annotations(split / self.METADATA_FILENAME)
                except Exception:
                    raise ValueError(f"Could not read annotation file {split / self.METADATA_FILENAME}")

                for dataset_piece in dataset_pieces:
                    item_metadata = {}
                    for k in dataset_piece.keys():
                        if (
                            k not in self.views_schema
                            and k not in self.entities_schema
                            and k not in self.annotations_schema
                        ):
                            item_metadata.update({k: dataset_piece.get(k, None)})
                    for k in item_metadata.keys():
                        dataset_piece.pop(k, None)

                    # create item
                    item = self._create_item(split.name, **item_metadata)

                    views_data: list[tuple[str, View]] = []
                    all_entities_data: dict[str, list[Entity]] = defaultdict(list)
                    all_annotations_data: dict[str, list[Annotation]] = defaultdict(list)
                    for k, v in dataset_piece.items():
                        if k in self.views_schema:
                            view_name = k
                            s = self.views_schema.get(view_name)
                            if s is None:
                                raise ValueError("View schema must be defined in the schemas argument.")
                            view_schema: type[View] = s

                            # create views
                            if isinstance(v, list):
                                if len(v) > 1:
                                    # create a mosaic from item images
                                    mosaic_file = mosaic(self.source_dir, split.name, v, view_name)
                                    view_file = self.source_dir / mosaic_file
                                else:
                                    view_file = self.source_dir / Path(v[0])
                                    if not view_file.is_file():  # no split path in metadata.jsonl
                                        view_file = self.source_dir / split / Path(v[0])
                                if view_file.is_file() and view_file.suffix in self.EXTENSIONS:
                                    view = self._create_vqa_view(item, view_file, view_schema)
                                    views_data.append((view_name, view))

                    for k, v in dataset_piece.items():
                        if k in self.entities_schema:
                            if source_id is None:
                                source_id = self.add_source("Builder", SourceKind.OTHER)
                            entity_name = k
                            raw_entities_data = v

                            # create entities and their annotations
                            entities_data, annotations_data = self._create_vqa_entities(
                                item, views_data, entity_name, raw_entities_data, source_id
                            )

                            for name, entities in entities_data.items():
                                all_entities_data[name].extend(entities)

                            for name, annotations in annotations_data.items():
                                all_annotations_data[name].extend(annotations)

                    yield {self.item_schema_name: item}
                    for view_name, view in views_data:
                        yield {view_name: view}

                    if all_entities_data is None:
                        continue

                    yield all_entities_data
                    yield all_annotations_data

    def _create_vqa_view(self, item: Item, view_file: Path, view_schema: type[View]) -> View:
        if not issubclass(view_schema, View):
            raise ValueError("View schema must be a subclass of View")
        if view_schema not in _PIXANO_SCHEMA_REGISTRY.values():
            raise ValueError(
                f"View schema {view_schema} is not supported. You should implement your own _create_view method."
            )

        view = create_instance_of_schema(
            view_schema,
            id=shortuuid.uuid(),
            item_ref=ItemRef(id=item.id),
            url=view_file,
            url_relative_path=self.source_dir,
        )
        view.url = str(self.url_prefix / view.url)
        return view

    def _create_vqa_entities(
        self,
        item: Item,
        views_data: list[tuple[str, View]],
        entity_name: str,
        raw_entities_data: list[Any],
        source_id: str,
    ) -> tuple[dict[str, list[Entity]], dict[str, list[Annotation]]]:
        def update_viewref(content, views_data, view_ref):
            match = re.match(r".*<image (\d)>.*", content)
            if match is None:
                return view_ref
            for m in match.groups():
                num = int(m) - 1
                if num >= 0 and num < len(views_data):
                    view_name, view = views_data[num]
                    return ViewRef(id=view.id, name=view_name)
            return view_ref

        entities: dict[str, list[Entity]] = defaultdict(list)
        annotations: dict[str, list[Annotation]] = defaultdict(list)

        s = self.entities_schema.get(entity_name)
        if s is None:
            raise ValueError("Entity schema must be defined in the schemas argument.")
        entity_schema: type[Entity] = s

        default_user = "import"
        if is_conversation(entity_schema):
            entities_data: list[list[Any]] = []

            if isinstance(raw_entities_data, list):
                if all(isinstance(data, list) for data in raw_entities_data):
                    entities_data = raw_entities_data
                if all(isinstance(data, dict) for data in raw_entities_data):
                    entities_data = [raw_entities_data]

            for conv in entities_data:
                view_name, view = views_data[0]
                default_view_ref = ViewRef(id=view.id, name=view_name)

                conversation = entity_schema(
                    id=shortuuid.uuid(), kind="vqa", item_ref=ItemRef(id=item.id), view_ref=default_view_ref
                )
                entities[entity_name].append(conversation)

                for i, q_n_a in enumerate(conv):
                    question = q_n_a.get("question")
                    if not question:
                        raise ValueError(f"No question found for item {item}")

                    # Handle question
                    if "user" not in question:
                        question["user"] = default_user

                    content = question.get("content")
                    if not content:
                        raise ValueError(f"No text content found for question in item {item}")

                    view_ref = update_viewref(content, views_data, default_view_ref)

                    query_msg = Message(
                        id=shortuuid.uuid(),
                        number=i,
                        type="QUESTION",
                        item_ref=ItemRef(id=item.id),
                        entity_ref=EntityRef(id=conversation.id, name=entity_name),
                        source_ref=SourceRef(id=source_id),
                        view_ref=view_ref,
                        timestamp=datetime.now(),
                        **question,
                    )
                    annotations["messages"].append(query_msg)

                    # Handle all responses

                    responses = []
                    response = q_n_a.get("response", None)
                    if isinstance(response, dict):
                        responses = [response]
                    if not responses:
                        responses = q_n_a.get("responses", [])
                        if not isinstance(responses, list):
                            responses = [responses]

                    for response in responses:
                        if "user" not in response:
                            response["user"] = default_user
                        content = response.get("content", "")
                        explanation = response.get("explanation", "")
                        if query_msg.question_type != "OPEN":
                            response["content"] = "[[{}]] {}".format(content, explanation)

                        response_msg = Message(
                            id=shortuuid.uuid(),
                            number=i,
                            type="ANSWER",
                            question_type=query_msg.question_type,
                            item_ref=ItemRef(id=item.id),
                            entity_ref=EntityRef(id=conversation.id, name=entity_name),
                            source_ref=SourceRef(id=source_id),
                            timestamp=datetime.now(),
                            **response,
                        )
                        annotations["messages"].append(response_msg)

        else:  # classic entity
            pass

        return entities, annotations

    def _read_annotations(self, annotations_file: Path) -> list[dict]:
        if not annotations_file.exists():
            raise FileNotFoundError(f"Anotations file {annotations_file} not found")
        return pa_json.read_json(annotations_file).to_pylist()



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from pixano.datasets.dataset_schema import DatasetItem
from pixano.datasets.workspaces import WorkspaceType
from pixano.datasets.workspaces.dataset_items import DefaultImageDatasetItem

from .base import FolderBaseBuilder


# Image extensions supported by Pillow
IMAGE_EXTENSIONS = [
    ".blp",
    ".bmp",
    ".dib",
    ".bufr",
    ".cur",
    ".pcx",
    ".dcx",
    ".dds",
    ".ps",
    ".eps",
    ".fit",
    ".fits",
    ".fli",
    ".flc",
    ".ftc",
    ".ftu",
    ".gbr",
    ".gif",
    ".grib",
    ".h5",
    ".hdf",
    ".png",
    ".apng",
    ".jp2",
    ".j2k",
    ".jpc",
    ".jpf",
    ".jpx",
    ".j2c",
    ".icns",
    ".ico",
    ".im",
    ".iim",
    ".tif",
    ".tiff",
    ".jfif",
    ".jpe",
    ".jpg",
    ".jpeg",
    ".mpg",
    ".mpeg",
    ".msp",
    ".pcd",
    ".pxr",
    ".pbm",
    ".pgm",
    ".ppm",
    ".pnm",
    ".psd",
    ".bw",
    ".rgb",
    ".rgba",
    ".sgi",
    ".ras",
    ".tga",
    ".icb",
    ".vda",
    ".vst",
    ".webp",
    ".wmf",
    ".emf",
    ".xbm",
    ".xpm",
]


class ImageFolderBuilder(FolderBaseBuilder):
    """Builder for image datasets stored in a folder."""

    EXTENSIONS = IMAGE_EXTENSIONS
    WORKSPACE_TYPE = WorkspaceType.IMAGE
    DEFAULT_SCHEMA: type[DatasetItem] = DefaultImageDatasetItem



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from typing import Annotated

import pandas as pd
from fastapi import APIRouter, Depends, HTTPException, Query

from pixano.app.models import ItemInfoModel, ItemModel
from pixano.app.settings import Settings, get_settings
from pixano.datasets.queries import TableQueryBuilder
from pixano.datasets.utils import DatasetAccessError, DatasetPaginationError
from pixano.features.schemas.schema_group import SchemaGroup
from pixano.utils.python import to_sql_list

from .utils import (
    assert_table_in_group,
    get_dataset,
    get_models_from_rows,
    get_rows,
)


router = APIRouter(prefix="/items_info", tags=["Items"])


@router.get("/{dataset_id}/", response_model=list[ItemInfoModel])
async def get_items_info(
    dataset_id: str,
    settings: Annotated[Settings, Depends(get_settings)],
    where: str | None = None,
    ids: list[str] | None = Query(None),
    limit: int | None = None,
    skip: int = 0,
) -> list[ItemInfoModel]:
    """Get items info.

    Args:
        dataset_id: Dataset ID.
        settings: App settings.
        where: Where clause for the item table.
        ids: IDs.
        limit: Limit number of items.
        skip: Skip number of items.

    Returns:
        List of items info.
    """
    dataset = get_dataset(dataset_id, settings.library_dir, None)
    assert_table_in_group(dataset, SchemaGroup.ITEM.value, SchemaGroup.ITEM)

    try:
        item_rows = get_rows(dataset, SchemaGroup.ITEM.value, where, ids, None, limit, skip)
    except DatasetPaginationError as err:
        raise HTTPException(status_code=400, detail=str(err))
    except DatasetAccessError as err:
        raise HTTPException(status_code=500, detail=str(err))

    items_models = get_models_from_rows(SchemaGroup.ITEM.value, ItemModel, item_rows)
    item_models_identified = {item.id: item for item in items_models}

    set_ids = {item.id for item in items_models}
    infos = {
        id: {
            group.value: {table: {"count": 0} for table in tables}
            for group, tables in dataset.schema.groups.items()
            if group.value not in [SchemaGroup.EMBEDDING.value, SchemaGroup.ITEM.value]
        }
        for id in set_ids
    }

    for table_name, table in dataset.open_tables().items():
        group_name = dataset.schema.get_table_group(table_name).value
        if group_name in [SchemaGroup.EMBEDDING.value, SchemaGroup.ITEM.value]:
            continue
        sql_ids = to_sql_list(set_ids)
        df: pd.DataFrame = (
            TableQueryBuilder(table).select(["item_ref.id"]).where(f"item_ref.id in {sql_ids}").to_pandas()
        )
        for id, count in df["item_ref.id"].value_counts().to_dict().items():
            infos[id][group_name][table_name] = {"count": count}

    items_info = [ItemInfoModel(info=info, **item_models_identified[id].model_dump()) for id, info in infos.items()]

    return items_info


@router.get("/{dataset_id}/{id}", response_model=ItemInfoModel)
async def get_item_info(
    dataset_id: str, id: str, settings: Annotated[Settings, Depends(get_settings)]
) -> ItemInfoModel:
    """Get an item info.

    Args:
        dataset_id: Dataset ID.
        id: ID.
        settings: App settings.

    Returns:
        The item info.
    """
    items_info = await get_items_info(dataset_id=dataset_id, settings=settings, ids=[id], limit=None, skip=0)

    return items_info[0]



# =====================================
# Copyright: CEA-LIST/DIASI/SIALV/LVA
# Author : pixano@cea.fr
# License: CECILL-C
# =====================================

from datetime import datetime


def issubclass_strict(obj: type, cls: type, strict: bool = False) -> bool:
    """Check if the given object is of the given class type or a subclass of the given class type.

    Args:
        obj: The object to check.
        cls: The class to compare against.
        strict: If True, the object must be of the given class type.

    Returns:
        True if the object is of the given class type or a subclass of the given class type.
    """
    if strict:
        return obj == cls
    return issubclass(obj, cls)


def validate_and_init_create_at_and_update_at(
    created_at: datetime | str | None, updated_at: datetime | str | None
) -> tuple[datetime, datetime]:
    """Validate and initialize created_at and updated_at.

    The validation and initialization of created_at and updated_at is done as follows:
    - If created_at is None, it is set to the current date and time.
    - If updated_at is None, it is set to created_at.
    - If updated_at is not None and created_at is None, a ValueError is raised.
    - If updated_at is not None and created_at is not None, updated_at should be greater than created_at.
    - If created_at and updated_at are provided as strings, they are converted to datetime objects from ISO format.

    Args:
        created_at: The creation date of the object.
        updated_at: The last modification date of the object.

    Returns:
        A tuple containing the created_at and updated_at.
    """
    if created_at is not None and not isinstance(created_at, datetime):
        if isinstance(created_at, str):
            created_at = datetime.fromisoformat(created_at)
        else:
            raise ValueError("created_at should be a datetime object or None.")
    if updated_at is not None and not isinstance(updated_at, datetime):
        if isinstance(updated_at, str):
            updated_at = datetime.fromisoformat(updated_at)
        else:
            raise ValueError("updated_at should be a datetime object or None.")
    if updated_at is not None and created_at is None:
        raise ValueError("created_at should be set if updated_at is set.")
    elif created_at is not None:
        if updated_at is not None:
            if not isinstance(updated_at, datetime):
                raise ValueError("updated_at should be a datetime object.")
            elif updated_at < created_at:
                raise ValueError("updated_at should be greater than created_at.")
        else:
            updated_at = created_at
    elif created_at is None:
        created_at = datetime.now()
        updated_at = created_at
    return created_at, updated_at
